# -*- coding: utf-8 -*-

#  Copyright 2019-  DNB
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

# For Python 2
from __future__ import unicode_literals
from __future__ import division
from io import open
from .compat import IS_PYTHON_2, STRING_TYPES

from pytz import utc
from tzlocal import get_localzone

from collections import OrderedDict
from copy import deepcopy
from datetime import datetime
from json import dumps, loads
from os import path, getcwd

from flex.core import validate_api_call
from genson import SchemaBuilder
from jsonpath_ng.ext import parse as parse_jsonpath
from jsonschema import validate, FormatChecker
from jsonschema.exceptions import SchemaError, ValidationError
from requests import request as client
from requests.exceptions import SSLError, Timeout

if IS_PYTHON_2:
    from urlparse import parse_qsl, urljoin, urlparse
else:
    from urllib.parse import parse_qsl, urljoin, urlparse

from robot.api import logger
from robot.api.deco import keyword
from robot.libraries.BuiltIn import BuiltIn

import os.path
import json
from jsonpath_rw import Index, Fields
from jsonpath_rw_ext import parse

from .schema_keywords import SCHEMA_KEYWORDS


class Keywords(object):

    def get_keyword_names(self):
        return [name for name in dir(self) if hasattr(getattr(self, name),
                                                      'robot_name')]

    ### Keywords start here

    

#    @keyword(name=None, tags=("Settings",))
#    def set_headers(self, headers):
#        """*Sets new request headers or updates the existing.*

#        ``headers``: The headers to add or update as a JSON object or a
#        dictionary.

#        *Examples*

#        | `Set Headers` | { "authorization": "Basic QWxhZGRpbjpPcGVuU2VzYW1"} |
#        | `Set Headers` | { "Accept-Encoding": "identity"} |
#        | `Set Headers` | ${auth_dict} |
#        """
#        self.request['headers'].update(self._input_object(headers))
#        return self.request['headers']

    

    

    

    

    @keyword(name=None, tags=("Requests",))
    def head(self, endpoint, headers=None, timeout=None, allow_redirects=None, validate=True):
        """*Sends a HEAD request to the endpoint.*

        *Arguments*

        ``endpoint``: REST API service endpoint URL - including resource path & query/path parameter(s) 

        ``headers``: Headers as a JSON object to add or override for the request.

        Following keywords can be used :  
        - `Load Headers` to load default headers
        - `Add Headers` to add new headers
        - `Update Header` to update value of existing header

        ``timeout``: Number of seconds to wait for the response before failing the keyword. No timeout set by default. For ex: timeout=2.5 

        ``allow_redirects``: If true, follow all redirects.
        In contrary to other methods, no HEAD redirects are followed by default.

        ``validate``: If false, skips any request and response validations set by expectation keywords.

        *Examples*

        Examples of `GET` will apply for `HEAD`.

        """
        endpoint = self._input_string(endpoint)
        request = deepcopy(self.request)
        request['method'] = "HEAD"
        if allow_redirects is not None:
            request['allowRedirects'] = self._input_boolean(allow_redirects)
        if timeout is not None:
            request['timeout'] = self._input_timeout(timeout)
        validate = self._input_boolean(validate)
        if headers:
            request['headers'].update(self._input_object(headers))
        return self._request(endpoint, request, validate)['response']

    @keyword(name=None, tags=("Requests",))
    def options(self, endpoint, headers=None, timeout=None, allow_redirects=None, validate=True):
        """*Sends an OPTIONS request to the endpoint.*

        *Arguments*

        ``endpoint``: REST API service endpoint URL - including resource path & query/path parameter(s)

        ``headers``: Headers as a JSON object to add or override for the request.

        Following keywords can be used :  
        - `Load Headers` to load default headers
        - `Add Headers` to add new headers
        - `Update Header` to update value of existing header

        ``timeout``: Number of seconds to wait for the response before failing the keyword. No timeout set by default. For ex: timeout=2.5

        ``allow_redirects``: If false, do not follow any redirects.

        ``validate``: If false, skips any request and response validations set by expectation keywords.

        *Examples*

        Examples of `GET` will apply for `OPTIONS`.
        """
        endpoint = self._input_string(endpoint)
        request = deepcopy(self.request)
        request['method'] = "OPTIONS"
        if allow_redirects is not None:
            request['allowRedirects'] = self._input_boolean(allow_redirects)
        if timeout is not None:
            request['timeout'] = self._input_timeout(timeout)
        validate = self._input_boolean(validate)
        if headers:
            request['headers'].update(self._input_object(headers))
        return self._request(endpoint, request, validate)['response']

    @keyword(name=None, tags=("Requests",))
    def get(self, endpoint, headers=None, query=None, timeout=None, allow_redirects=None, validate=True):
        """*Sends a GET request to the endpoint.*

        *Arguments*

        ``endpoint``: REST API service endpoint URL - including resource path & query/path parameter(s)

        ``headers``: Headers as a JSON object to add or override for the request.

        Following keywords can be used :  
        - `Load Headers` to load default headers
        - `Add Headers` to add new headers
        - `Update Header` to update value of existing header

        ``query``: Request query parameters as a JSON object or a dictionary.
        Alternatively, query parameters can be given as part of endpoint as well.

        ``timeout``: Number of seconds to wait for the response before failing the keyword. No timeout set by default. For ex: timeout=2.5

        ``allow_redirects``: If false, do not follow any redirects.

        ``validate``: If false, skips any request and response validations set by expectation keywords.

        *Examples*

		``Set Headers:``
        | `GET` | https://jsonplaceholder.typicode.com/users   | headers={ "Authentication": "Basic ..", "Trace-Id": "1234" } | timeout=2.5 |

        ``Load Headers from JSON File:``
        | ${requestHeaders} |  `Load Headers`  | ${EXECDIR}/path/to/file.json |
        | `GET` | https://jsonplaceholder.typicode.com/users   | headers=${requestHeaders} | 

        ``Load Headers (as a template) from JSON file and Update values of headers:``
        | ${requestHeaders} |  `Load Headers`  | ${EXECDIR}/path/to/file.json |
        | ${updatedHeaders} |  `Update Header` | ${requestHeaders} |  X-DNBAPI-UserId  |  TBxxxxx |
        | `GET` | https://jsonplaceholder.typicode.com/users   | headers=${updatedHeaders} |

        ``Query Params:``       
        | `GET` | https://jsonplaceholder.typicode.com/users?_limit=2 |
        | `GET` | https://jsonplaceholder.typicode.com/users   | query={ "_limit": "2" } | #same as above |            

        ``Set Query Params :``
        | ${updatedEndpoint} | `Set Query Param` | https://testserver/v1/accounts | accountNumber | 12085214800 |
        | `GET` | ${updatedEndpoint} |

        ``Path Params:``
        | `GET` | https://jsonplaceholder.typicode.com/users/1   |

        ``Set Path Params:``
        | ${updatedEndpoint} | `Set Path Param` | https://jsonplaceholder.typicode.com/users/{id} | {id} | 1 |
        | `GET` | ${updatedEndpoint} |  
        """
        endpoint = self._input_string(endpoint)
        request = deepcopy(self.request)
        request['method'] = "GET"
        request['query'] = OrderedDict()
        query_in_url = OrderedDict(parse_qsl(urlparse(endpoint).query))
        if query_in_url:
            request['query'].update(query_in_url)
            endpoint = endpoint.rsplit('?', 1)[0]
        if query:
            request['query'].update(self._input_object(query))
        if allow_redirects is not None:
            request['allowRedirects'] = self._input_boolean(allow_redirects)
        if timeout is not None:
            request['timeout'] = self._input_timeout(timeout)
        validate = self._input_boolean(validate)
        if headers:
            request['headers'].update(self._input_object(headers))
        return self._request(endpoint, request, validate)['response']
        
        
    @keyword(name=None, tags=("Requests",))
    def post(self, endpoint, body=None, timeout=None, allow_redirects=None, validate=True, headers=None):
        """*Sends a POST request to the endpoint.*

        *Arguments*

        ``endpoint``: REST API service endpoint URL - including resource path & query/path parameter(s)

        ``headers``: Headers as a JSON object to add or override for the request.

        Following keywords can be used :  
        - `Load Headers` to load default headers
        - `Add Headers` to add new headers
        - `Update Header` to update value of existing header

        ``body``: Request body as a JSON object.

        Following keywords can be used :  
        - `Load Payload` to load default payload
        - `Update Payload` to update value of existing payload

        ``timeout``: Number of seconds to wait for the response before failing the keyword. No timeout set by default. For ex: timeout=2.5

        ``allow_redirects``: If false, do not follow any redirects.

        ``validate``: If false, skips any request and response validations set by expectation keywords.

        *Examples*

        Examples of `GET` will apply for `POST`. Below examples show usage of payload keywords

        ``POST request by passing payload manually``
        | `POST` | https://jsonplaceholder.typicode.com/users   | body={ "id": 11, "name": "Gil Alexander" } | 

        ``POST request by loading payload from a JSON file``
        | ${requestPayload} |  `Load Payload`  | ${EXECDIR}/path/to/file.json |
        | `POST` | https://jsonplaceholder.typicode.com/users   | body=${requestPayload} | 

        ``POST request by loading payload (as a template) from a JSON file and updating values of payload``
        | ${requestPayload} |  `Load Payload`  | ${EXECDIR}/path/to/file.json |
        | ${updatedPayload} |  `Update Payload` | ${requestPayload} |  name  |  Tom Cruise |
        | `POST` | https://jsonplaceholder.typicode.com/users   | body=${updatedPayload} |
        """
        endpoint = self._input_string(endpoint)
        request = deepcopy(self.request)
        request['method'] = "POST"
        request['body'] = self.input(body)
        if allow_redirects is not None:
            request['allowRedirects'] = self._input_boolean(allow_redirects)
        if timeout is not None:
            request['timeout'] = self._input_timeout(timeout)
        validate = self._input_boolean(validate)
        if headers:
            request['headers'].update(self._input_object(headers))
        return self._request(endpoint, request, validate)['response']

    @keyword(name=None, tags=("Requests",))
    def put(self, endpoint, headers=None, body=None, timeout=None, allow_redirects=None, validate=True):
        """*Sends a PUT request to the endpoint.*

        *Arguments*

        ``endpoint``: REST API service endpoint URL - including resource path & query/path parameter(s)

        ``headers``: Headers as a JSON object to add or override for the request.

        Following keywords can be used :  
        - `Load Headers` to load default headers
        - `Add Headers` to add new headers
        - `Update Header` to update value of existing header

        ``body``: Request body as a JSON object.

        Following keywords can be used :  
        - `Load Payload` to load default payload
        - `Update Payload` to update value of existing payload

        ``timeout``: Number of seconds to wait for the response before failing the keyword. No timeout set by default. For ex: timeout=2.5

        ``allow_redirects``: If false, do not follow any redirects.

        ``validate``: If false, skips any request and response validations set by expectation keywords.

        *Examples*

        Examples of `GET` & `POST` will apply for `PUT`


        """
        endpoint = self._input_string(endpoint)
        request = deepcopy(self.request)
        request['method'] = "PUT"
        request['body'] = self.input(body)
        if allow_redirects is not None:
            request['allowRedirects'] = self._input_boolean(allow_redirects)
        if timeout is not None:
            request['timeout'] = self._input_timeout(timeout)
        validate = self._input_boolean(validate)
        if headers:
            request['headers'].update(self._input_object(headers))
        return self._request(endpoint, request, validate)['response']

    @keyword(name=None, tags=("Requests",))
    def patch(self, endpoint, headers=None, body=None, timeout=None, allow_redirects=None, validate=True):
        """*Sends a PATCH request to the endpoint.*

        *Arguments*

        ``endpoint``: REST API service endpoint URL - including resource path & query/path parameter(s)

        ``headers``: Headers as a JSON object to add or override for the request.

        Following keywords can be used :  
        - `Load Headers` to load default headers
        - `Add Headers` to add new headers
        - `Update Header` to update value of existing header

        ``body``: Request body as a JSON object.

        Following keywords can be used :  
        - `Load Payload` to load default payload
        - `Update Payload` to update value of existing payload

        ``timeout``: Number of seconds to wait for the response before failing the keyword. No timeout set by default. For ex: timeout=2.5

        ``allow_redirects``: If false, do not follow any redirects.

        ``validate``: If false, skips any request and response validations set by expectation keywords.

        *Examples*

        Examples of `GET` & `POST` will apply for `PATCH`

        """
        endpoint = self._input_string(endpoint)
        request = deepcopy(self.request)
        request['method'] = "PATCH"
        request['body'] = self.input(body)
        if allow_redirects is not None:
            request['allowRedirects'] = self._input_boolean(allow_redirects)
        if timeout is not None:
            request['timeout'] = self._input_timeout(timeout)
        validate = self._input_boolean(validate)
        if headers:
            request['headers'].update(self._input_object(headers))
        return self._request(endpoint, request, validate)['response']

    @keyword(name=None, tags=("Requests",))
    def delete(self, endpoint, headers=None, timeout=None, allow_redirects=None, validate=True):
        """*Sends a DELETE request to the endpoint.*

        *Arguments*

        ``endpoint``: REST API service endpoint URL - including resource path & query/path parameter(s)

        ``headers``: Headers as a JSON object to add or override for the request.

        Following keywords can be used :  
        - `Load Headers` to load default headers
        - `Add Headers` to add new headers
        - `Update Header` to update value of existing header

        ``timeout``: Number of seconds to wait for the response before failing the keyword. No timeout set by default. For ex: timeout=2.5

        ``allow_redirects``: If false, do not follow any redirects.

        ``validate``: If false, skips any request and response validations set by expectation keywords.

        *Examples*

        Examples of `GET` will apply for `DELETE`


        """
        endpoint = self._input_string(endpoint)
        request = deepcopy(self.request)
        request['method'] = "DELETE"
        if allow_redirects is not None:
            request['allowRedirects'] = self._input_boolean(allow_redirects)
        if timeout is not None:
            request['timeout'] = self._input_timeout(timeout)
        validate = self._input_boolean(validate)
        if headers:
            request['headers'].update(self._input_object(headers))
        return self._request(endpoint, request, validate)['response']

    
    
    def input(self, what):
        """*Converts the input to JSON and returns it.*
        Any of the following is accepted:
        - The path to JSON file
        - Any scalar that can be interpreted as JSON
        - A dictionary or a list
        *Examples*
        | ${payload} | `Input` | ${CURDIR}/payload.json |
        | ${object} | `Input` | { "name": "Julie Langford", "username": "jlangfor" } |
        | ${object} | `Input` | ${dict} |
        | ${array} | `Input` | ["name", "username"] |
        | ${array} | `Input` | ${list} |
        | ${boolean} | `Input` | true |
        | ${boolean} | `Input` | ${True} |
        | ${number} | `Input` | 2.0 |
        | ${number} | `Input` | ${2.0} |
        | ${string} | `Input` | Quotes are optional for strings |
        """
        if what is None:
            return None
        if not isinstance(what, STRING_TYPES):
            return self._input_json_from_non_string(what)
        if path.isfile(what):
            return self._input_json_from_file(what)
        try:
            return self._input_json_as_string(what)
        except ValueError:
            return self._input_string(what)    
        
    
    @keyword(name=None, tags=("Assertions",))
    def assert_if_field_is_not_found(self, field, response=None):
        """*Asserts that the JSON field does not exist. Throws Assertion error if field is found*

        *Arguments*

        ``field``: 

        The field consists of parts separated by spaces, the parts being
        object property names or array indices starting from 0, and the root
        being the instance created by the last request (see `Output` for it).

        For asserting deeply nested properties or multiple objects at once,
        [http://goessner.net/articles/JsonPath|JSONPath] can be used with
        the root being the response body.

		``response``: Response object for validation 

        *Examples*

        | `GET`    | https://jsonplaceholder.typicode.com/users/1 |
        | `Assert If Field is Not Found` | response body password |
        | `Assert If Field is Not Found` | $.password |
        | `Assert If Field is Not Found` | $..geo.elevation | # JSONPath represention |
        | `Assert If Field is Not Found` | response body address geo elevation | # Same as above, fields seperated by spaces |

        | `GET`    | https://jsonplaceholder.typicode.com/users/1   |
        | `Assert If Field is Not Found` | response body 0 password |
        | `Assert If Field is Not Found` | $[*].password |
        | `Assert If Field is Not Found` | $[*]..geo.elevation |
        """
        try:
            matches = self._find_by_field(field, response, print_found=False)
        except AssertionError:
            return
        for found in matches:
            self.log_json(found['reality'],
                          "\n\nExpected '%s' to not exist, but it is:" % (field))
        raise AssertionError("Expected '%s' to not exist, but it does." % (
            field))

    @keyword(name=None, tags=("Assertions",))
    def assert_if_field_is_null(self, field, response=None, **validations):
        """*Asserts that the JSON field is null. Throws error if field is not null*

        *Arguments*

        ``field``: 

        The field consists of parts separated by spaces, the parts being
        object property names or array indices starting from 0, and the root
        being the instance created by the last request (see `Output` for it).

        For asserting deeply nested properties or multiple objects at once,
        [http://goessner.net/articles/JsonPath|JSONPath] can be used with
        the root being the response body.

        ``response``: Response object for validation 
        
        ``validations``:

        The JSON Schema validation keywords
        [https://json-schema.org/understanding-json-schema/reference/generic.html|common for all types]
        can be used. Validations are optional but update the schema of
        the property (more accurate) if given.
        `Output Schema` can be used for the current schema of the field.

        The keyword will fail if any of the given validations fail.
        Given validations can be skipped altogether by adding ``skip=true``.
        When skipped, the schema is updated but the validations are not ran.
        Skip is intented mainly for debugging the updated schema before aborting.

        *Examples*

        | `PUT`  		   | https://jsonplaceholder.typicode.com/users/1 | { "deactivated_at": null } | |
        | `Assert If Field is Null` | response body deactivated_at | # JSONPath alternative |
        | `Assert If Field is Null` | $.deactivated_at | | |
        """
        values = []
        for found in self._find_by_field(field, response):
            reality = found['reality']
            schema = {"type": "null"}
            skip = self._input_boolean(validations.pop('skip', False))
            if not skip:
                self._assert_schema(schema, reality)
            values.append(reality)
        return values

    @keyword(name=None, tags=("Assertions",))
    def assert_if_field_is_boolean(self, field, response=None, value=None, **validations):
        """*Asserts that the JSON field is boolean.*

        *Arguments*

        ``field``: 

        The field consists of parts separated by spaces, the parts being
        object property names or array indices starting from 0, and the root
        being the instance created by the last request (see `Output` for it).

        For asserting deeply nested properties or multiple objects at once,
        [http://goessner.net/articles/JsonPath|JSONPath] can be used with
        the root being the response body.

        ``response``: Response object for validation 
        
        ``value``:

        If given, the property value is validated in addition to the type.

        ``validations``:

        The JSON Schema validation keywords
        [https://json-schema.org/understanding-json-schema/reference/generic.html|common for all types]
        can be used. Validations are optional but update the schema of
        the property (more accurate) if given.
        `Output Schema` can be used for the current schema of the field.

        The keyword will fail if any of the given validations fail.
        Given validations can be skipped altogether by adding ``skip=true``.
        When skipped, the schema is updated but the validations are not ran.
        Skip is intented mainly for debugging the updated schema before aborting.

        *Examples*

        | `PUT`  		      | https://jsonplaceholder.typicode.com/users/1 | { "verified_email": true } | | |  |
        | `Assert If Field is Boolean` | response body verified_email | | | | # value is optional |
        | `Assert If Field is Boolean` | response body verified_email | true | | | # JSONPath alternative |
        | `Assert If Field is Boolean` | response body verified_email | ${True} | | | # same as above |
        | `Assert If Field is Boolean` | $.verified_email | true | | | |
        | `Assert If Field is Boolean` | $.verified_email | true | enum=[1, "1"] | skip=true | # would pass |
        """
        values = []
        for found in self._find_by_field(field, response):
            reality = found['reality']
            schema = {"type": "boolean"}
            if value is not None:
                schema['enum'] = [self._input_boolean(value)]
            elif self._should_add_examples():
                schema['examples'] = [reality]
            skip = self._input_boolean(validations.pop('skip', False))
            if not skip:
                self._assert_schema(schema, reality)
            values.append(reality)
        return values

    @keyword(name=None, tags=("Assertions",))
    def assert_if_field_is_integer(self, field, *enum, response=None, **validations):
        """*Asserts that the JSON field is Integer.*

        *Arguments*

        ``field``: 

        The field consists of parts separated by spaces, the parts being
        object property names or array indices starting from 0, and the root
        being the instance created by the last request (see `Output` for it).

        For asserting deeply nested properties or multiple objects at once,
        [http://goessner.net/articles/JsonPath|JSONPath] can be used with
        the root being the response body.

        ``response``: Response object for validation 
        
        ``enum``:

        The allowed values for the property as zero or more arguments.
        If none given, the value of the property is not asserted.

        ``validations``:

        The JSON Schema validation keywords
        [https://json-schema.org/understanding-json-schema/reference/numeric.html#integer|for numeric types]
        can be used. Validations are optional but update the schema of
        the property (more accurate) if given.
        `Output Schema` can be used for the current schema of the field.

        The keyword will fail if any of the given validations fail.
        Given validations can be skipped altogether by adding ``skip=true``.
        When skipped, the schema is updated but the validations are not ran.
        Skip is intented mainly for debugging the updated schema before aborting.

        *Examples*

        | `GET`  			  | https://jsonplaceholder.typicode.com/users/1 | |  |
        | `Assert If Field is Integer` | response body id | | # value is optional |
        | `Assert If Field is Integer` | response body id | 1 | # JSONPath alternative |
        | `Assert If Field is Integer` | response body id | ${1} | # same as above |
        | `Assert If Field is Integer` | $.id | 1 | |

        | `GET`  			  | https://jsonplaceholder.typicode.com/users?_limit=10 | | | | |
        | `Assert If Field is Integer` | response body 0 id | 1 | | |
        | `Assert If Field is Integer` | $[0].id | 1 | | | # same as above |
        | `Assert If Field is Integer` | $[*].id | | minimum=1 | maximum=10 |
        """
        values = []
        for found in self._find_by_field(field, response):
            schema = found['schema']
            reality = found['reality']
            skip = self._input_boolean(validations.pop('skip', False))
            self._set_type_validations("integer", schema, validations)
            if enum:
                if 'enum' not in schema:
                    schema['enum'] = []
                for value in enum:
                    value = self._input_integer(value)
                    if value not in schema['enum']:
                        schema['enum'].append(value)
            elif self._should_add_examples():
                schema['examples'] = [reality]
            if not skip:
                self._assert_schema(schema, reality)
            values.append(reality)
        return values

    @keyword(name=None, tags=("Assertions",))
    def assert_if_field_is_number(self, field, *enum, response=None, **validations):
        """*Asserts that the JSON field is number.*

        *Arguments*

        ``field``: 

        The field consists of parts separated by spaces, the parts being
        object property names or array indices starting from 0, and the root
        being the instance created by the last request (see `Output` for it).

        For asserting deeply nested properties or multiple objects at once,
        [http://goessner.net/articles/JsonPath|JSONPath] can be used with
        the root being the response body.

        ``response``: Response object for validation 
        
        ``enum``:

        The allowed values for the property as zero or more arguments.
        If none given, the value of the property is not asserted.

        ``validations``:

        The JSON Schema validation keywords
        [https://json-schema.org/understanding-json-schema/reference/numeric.html#number|for numeric types] can be used. Validations are optional but update the schema of
        the property (more accurate) if given.
        `Output Schema` can be used for the current schema of the field.

        The keyword will fail if any of the given validations fail.
        Given validations can be skipped altogether by adding ``skip=true``.
        When skipped, the schema is updated but the validations are not ran.
        Skip is intented mainly for debugging the updated schema before aborting.

        *Examples*

        | `PUT`  | https://jsonplaceholder.typicode.com/users/1 | { "allocation": 95.0 } | |
        | `Number` | response body allocation | | # value is optional |
        | `Number` | response body allocation | 95.0 | # JSONPath alternative |
        | `Number` | response body allocation | ${95.0} | # same as above |
        | `Number` | $.allocation | 95.0 | # JSONPath representation |

        | `GET`  | https://jsonplaceholder.typicode.com/users?_limit=10 | | | | |
        | `Number` | $[0].id | 1 | | | # integers are also numbers |
        | `Number` | $[*].id | | minimum=1 | maximum=10 |
        """
        values = []
        for found in self._find_by_field(field, response):
            schema = found['schema']
            reality = found['reality']
            skip = self._input_boolean(validations.pop('skip', False))
            self._set_type_validations("number", schema, validations)
            if enum:
                if 'enum' not in schema:
                    schema['enum'] = []
                for value in enum:
                    value = self._input_number(value)
                    if value not in schema['enum']:
                        schema['enum'].append(value)
            elif self._should_add_examples():
                schema['examples'] = [reality]
            if not skip:
                self._assert_schema(schema, reality)
            values.append(reality)
        return values
        

    @keyword(name=None, tags=("Assertions",))
    def assert_if_field_is_string(self, field, *enum, response=None, **validations):
        """*Asserts that the JSON field is String.*

        *Arguments*

        ``field``: 

        The field consists of parts separated by spaces, the parts being
        object property names or array indices starting from 0, and the root
        being the instance created by the last request (see `Output` for it).

        For asserting deeply nested properties or multiple objects at once,
        [http://goessner.net/articles/JsonPath|JSONPath] can be used with
        the root being the response body.

        ``response``: Response object for validation 
        
        ``enum``:

        The allowed values for the property as zero or more arguments.
        If none given, the value of the property is not asserted.

        ``validations``:

        The JSON Schema validation keywords
        [https://json-schema.org/understanding-json-schema/reference/string.html|for string]
        can be used. Validations are optional but update the schema of
        the property (more accurate) if given.
        `Output Schema` can be used for the current schema of the field.

        The keyword will fail if any of the given validations fail.
        Given validations can be skipped altogether by adding ``skip=true``.
        When skipped, the schema is updated but the validations are not ran.
        Skip is intented mainly for debugging the updated schema before aborting.
      
        
        *Examples*

        | `GET`  			 | https://jsonplaceholder.typicode.com/users/1 | | | |
        | `Assert If Field is String` | response body email | | | # value is optional |
        | `Assert If Field is String` | response body email | Sincere@april.biz |
        | `Assert If Field is String` | $.email | Sincere@april.biz | | |
        | `Assert If Field is String` | $.email | | format=email |

        | `GET`  			 | https://jsonplaceholder.typicode.com/users?_limit=10 | | | |
        | `Assert If Field is String` | response body 0 email | Sincere@april.biz |
        | `Assert If Field is String` | $[0].email | Sincere@april.biz | | # same as above |
        | `Assert If Field is String` | $[*].email | | format=email |
        """
        values = []
        for found in self._find_by_field(field, response):
            schema = found['schema']
            reality = found['reality']
            skip = self._input_boolean(validations.pop('skip', False))
            self._set_type_validations("string", schema, validations)
            if enum:
                if 'enum' not in schema:
                    schema['enum'] = []
                for value in enum:
                    value = self._input_string(value)
                    if value not in schema['enum']:
                        schema['enum'].append(value)
            elif self._should_add_examples():
                schema['examples'] = [reality]
            if not skip:
                self._assert_schema(schema, reality)
            values.append(reality)
        return values       
         

    @keyword(name=None, tags=("Assertions",))
    def assert_if_field_is_object(self, field, *enum, response=None, **validations):
        """*Asserts that the JSON field is Object.*

        *Arguments*

        ``field``: 

        The field consists of parts separated by spaces, the parts being
        object property names or array indices starting from 0, and the root
        being the instance created by the last request (see `Output` for it).

        For asserting deeply nested properties or multiple objects at once,
        [http://goessner.net/articles/JsonPath|JSONPath] can be used with
        the root being the response body.

        ``response``: Response object for validation 
        
        ``enum``:

        The allowed values for the property as zero or more arguments.
        If none given, the value of the property is not asserted.

        ``validations``:

        The JSON Schema validation keywords
        [https://json-schema.org/understanding-json-schema/reference/object.html|for object]
        can be used. Validations are optional but update the schema of
        the property (more accurate) if given.
        `Output Schema` can be used for the current schema of the field.

        The keyword will fail if any of the given validations fail.
        Given validations can be skipped altogether by adding ``skip=true``.
        When skipped, the schema is updated but the validations are not ran.
        Skip is intented mainly for debugging the updated schema before aborting.

        *Examples*

        | `GET`  			 | https://jsonplaceholder.typicode.com/users/1 | | |
        | `Assert If Field is Object` | response body | |
        | `Assert If Field is Object` | response body | required=["id", "name"] | # can have other properties |

        | `GET`  			 | https://jsonplaceholder.typicode.com/users/1 | | |
        | `Assert If Field is Object` | $.address.geo | required=["lat", "lng"] |
        | `Assert If Field is Object` | $..geo | additionalProperties=false | # cannot have other properties |
        """
        values = []
        for found in self._find_by_field(field, response):
            schema = found['schema']
            reality = found['reality']
            skip = self._input_boolean(validations.pop('skip', False))
            self._set_type_validations("object", schema, validations)
            if enum:
                if 'enum' not in schema:
                    schema['enum'] = []
                for value in enum:
                    value = self._input_object(value)
                    if value not in schema['enum']:
                        schema['enum'].append(value)
            elif self._should_add_examples():
                schema['examples'] = [reality]
            if not skip:
                self._assert_schema(schema, reality)
            values.append(reality)
        return values

    @keyword(name=None, tags=("Assertions",))
    def assert_if_field_is_array(self, field, *enum, response=None, **validations):
        """*Asserts the JSON field is Array.*

        *Arguments*

        ``field``: 

        The field consists of parts separated by spaces, the parts being
        object property names or array indices starting from 0, and the root
        being the instance created by the last request (see `Output` for it).

        For asserting deeply nested properties or multiple objects at once,
        [http://goessner.net/articles/JsonPath|JSONPath] can be used with
        the root being the response body.

        ``response``: Response object for validation 
        
        ``enum``:

        The allowed values for the property as zero or more arguments.
        If none given, the value of the property is not asserted.

        ``validations``:

        The JSON Schema validation keywords
        [https://json-schema.org/understanding-json-schema/reference/array.html|for array]
        can be used. Validations are optional but update the schema of
        the property (more accurate) if given.
        `Output Schema` can be used for the current schema of the field.

        The keyword will fail if any of the given validations fail.
        Given validations can be skipped altogether by adding ``skip=true``.
        When skipped, the schema is updated but the validations are not ran.
        Skip is intented mainly for debugging the updated schema before aborting.

        *Examples*
        | `GET`  			| https://jsonplaceholder.typicode.com/users?_limit=10 | | | |
        | `Assert If Field is Array` | response body | | |
        | `Assert If Field is Array` | $ | # same as above | 
        | `Assert If Field is Array` | $ | minItems=1 | maxItems=10 | uniqueItems=true |
        """
        values = []
        for found in self._find_by_field(field, response):
            schema = found['schema']
            reality = found['reality']
            skip = self._input_boolean(validations.pop('skip', False))
            self._set_type_validations("array", schema, validations)
            if enum:
                if 'enum' not in schema:
                    schema['enum'] = []
                for value in enum:
                    value = self._input_array(value)
                    if value not in schema['enum']:
                        schema['enum'].append(value)
            elif self._should_add_examples():
                schema['examples'] = [reality]
            if not skip:
                self._assert_schema(schema, reality)
            values.append(reality)
        return values

#    @keyword(name=None, tags=("Settings",))
#    def convert_input_to_json(self, what):
#        """*Converts the input to JSON and returns it.*

#        Any of the following is accepted:

#        - The path to JSON file
#        - Any scalar that can be interpreted as JSON
#        - A dictionary or a list

#        *Examples*

#        | ${payload} | `Convert Input To JSON` | ${CURDIR}/payload.json |

#        | ${object} | `Convert Input To JSON` | { "name": "Julie Langford", "username": "jlangfor" } |
#        | ${object} | `Convert Input To JSON` | ${dict} |

#        | ${array} | `Convert Input To JSON` | ["name", "username"] |
#        | ${array} | `Convert Input To JSON` | ${list} |

#        | ${boolean} | `Convert Input To JSON` | true |
#        | ${boolean} | `Convert Input To JSON` | ${True} |

#        | ${number} | `Convert Input To JSON` | 2.0 |
#        | ${number} | `Convert Input To JSON` | ${2.0} |

#        | ${string} | `Convert Input To JSON` | Quotes are optional for strings |
#        """
#        if what is None:
#            return None
#        if not isinstance(what, STRING_TYPES):
#            return self._input_json_from_non_string(what)
#        if path.isfile(what):
#            return self._input_json_from_file(what)
#        try:
#            return self._input_json_as_string(what)
#        except ValueError:
#            return self._input_string(what)    

    

    @keyword(name=None, tags=("Logging",))
    def print(self, element="", file_path=None, append=False, sort_keys=False):
        """*Outputs the element as JSON to terminal or a file.*


        *Arguments*

        ``element``: By default, the last request and response are output to terminal.
        The output can be limited further by: 

        - The property of last instance, e.g. ``request`` or ``response``
        - Any nested property that exists, e.g. using JSONPath


        ``file_path``: The JSON is written to a file instead of terminal.
        The file is created if it does not exist.

        ``append``: If true, the JSON is appended to the given file
        instead of truncating it first.

        ``sort_keys``: If true, the JSON is sorted alphabetically by
        property names before it is output.

        *Examples*

        | `Print` | response | # only the response is output |
        | `Print` | response body | # only the response body is output |
        | `Print` | $.email | # only the response body property is output |
        | `Print` | $..geo | # only the nested response body property is output |

        | `Print` | request | # only the request is output |
        | `Print` | request headers | # only the request headers are output |
        | `Print` | request headers Authentication | # only this header is output |

        | `Print` | response body | ${CURDIR}/response_body.json | | # write the response body to a file |
        | `Print` | response seconds | ${CURDIR}/response_delays.log | append=true | # keep track of response delays in a file |
        """
        if isinstance(element, (STRING_TYPES)):
            if element == "":
                try:
                    json = deepcopy(self._last_instance_or_error())
                    json.pop('schema')
                    json.pop('spec')
                except IndexError:
                    raise RuntimeError(no_instances_error)
            elif element.startswith("schema"):
                logger.warn("Using `Output` for schema is deprecated. " +
                            "Using `Output Schema` to handle schema paths better.")
                element = element.lstrip("schema").lstrip()
                return self.output_schema(element, file_path, append, sort_keys)
            elif element.startswith(("request", "response", "$")):
                self._last_instance_or_error()
                matches = self._find_by_field(element, return_schema=False)
                if len(matches) > 1:
                    json = [found['reality'] for found in matches]
                else:
                    json = matches[0]['reality']
            else:
                try:
                    json = self._input_json_as_string(element)
                except ValueError:
                    json = self._input_string(element)
        else:
            json = self._input_json_from_non_string(element)
        sort_keys = self._input_boolean(sort_keys)
        if not file_path:
            self.log_json(json, sort_keys=sort_keys)
        else:
            content = dumps(json, ensure_ascii=False, indent=4,
                            separators=(',', ': '), sort_keys=sort_keys)
            write_mode = 'a' if self._input_boolean(append) else 'w'
            try:
                with open(path.join(getcwd(), file_path), write_mode,
                          encoding="utf-8") as file:
                    if IS_PYTHON_2:
                        content = unicode(content)
                    file.write(content)
            except IOError as e:
                raise RuntimeError("Error outputting to file '%s':\n%s" % (
                    file_path, e))
        return json

#    @keyword(name=None, tags=("Logging",))
#    def rest_instances(self, file_path=None, sort_keys=False):
#        """*Writes the instances as JSON to a file*

#        The instances are written to file as a JSON array of JSON objects,
#        each object representing a single instance, and having three properties:

#        - the request
#        - the response
#        - the schema for both, which have been updated according to the tests

#        The file is created if it does not exist, otherwise it is truncated.

#        *Arguments*

#        ``sort_keys``: If true, the instances are sorted alphabetically by
#        property names.

#        *Examples*

#        | `Rest Instances` | ${CURDIR}/log.json |
#        """
#        if not file_path:
#            outputdir_path = BuiltIn().get_variable_value("${OUTPUTDIR}")
#            if self.request['netloc']:
#                file_path = path.join(outputdir_path,
#                                      self.request['netloc']) + '.json'
#            else:
#                file_path = path.join(outputdir_path, "instances") + '.json'
#        sort_keys = self._input_boolean(sort_keys)
#        content = dumps(self.instances, ensure_ascii=False, indent=4,
#                        separators=(',', ': '), sort_keys=sort_keys)
#        try:
#            with open(file_path, 'w', encoding="utf-8") as file:
#                if IS_PYTHON_2:
#                    content = unicode(content)
#                file.write(content)
#        except IOError as e:
#            raise RuntimeError("Error exporting instances " +
#                               "to file '%s':\n%s" % (file_path, e))
#        return self.instances
        
    
    @keyword(name=None, tags=("Assertions",))
    def extract_response_for_assertions(self):
        """*Extracts the response into a variable that can further be used in assertions and value chain testing. Returns RESTInstance object that includes both request and response objects*

        *Examples*

        | ${response} | `Extract Response For Assertions` | 
        """
        return self.instances[-1]    

    

    @keyword(name=None, tags=("Settings",))
    def update_payload(self, request_payload, json_path, new_value):
        """*Update value to JSON payload using JSONPath. Returns updatedPayload as JSON object*

        *Arguments*

            ``request_payload``: json object

            ``json_path``: jsonpath expression

            ``new_value``: value to update

        *Examples*

        | ${updatedPayload}=  |  `Update Payload` | ${requestPayload} |  $..address.streetAddress  |  Ratchadapisek Road |
        """
        updated_payload = dict(request_payload)
        json_path_expr = parse(json_path)
        for match in json_path_expr.find(updated_payload):
            path = match.path
            if isinstance(path, Index):
                match.context.value[match.path.index] = new_value
            elif isinstance(path, Fields):
                match.context.value[match.path.fields[0]] = new_value
        return updated_payload

    @keyword(name=None, tags=("Settings",))
    def delete_header(self, request_headers, header_name):
        """*Delete a dictionary or list object from existing JSON header object. Returns updatedHeaders as JSON Object*

	    *Arguments*
	
	        ``request_headers``: json object
	
	        ``header_name``: Header to be deleted
	
	    *Examples*
	
	    | ${updatedHeaders}  |  `Delete Header` | ${requestHeaders} |  Authorization  |
	    """
        json_headers = self.input(request_headers)
        json_path_expr = parse('$' + '.' + header_name)
        matches = json_path_expr.find(json_headers)
        if not matches:
            raise AssertionError("Header '%s' " % (header_name) +
                             "did not match anything.")
        for match in json_path_expr.find(json_headers):
	        path = match.path
	        if isinstance(path, Index):
	            del (match.context.value[match.path.index])
	        elif isinstance(path, Fields):
	            del (match.context.value[match.path.fields[0]])
        return json_headers

    

    @keyword(name=None, tags=("Settings",))
    def load_headers(self, file_name):
        """*Load Request Headers as JSON file. Returns requestHeaders as JSON Object*

        *Arguments*

            ``file_name``: json file name (with absolute path)

        *Examples*

        | ${requestHeaders}=  |  `Load Headers`  | /path/to/file.json |
        """
        logger.debug("Check if file exists")
        if os.path.isfile(file_name) is False:
            logger.error("JSON file: " + file_name + " not found")
            raise IOError
        elif os.path.getsize(file_name) is 0:
            logger.error("JSON file: " + file_name + " is empty")
            raise IOError
        with open(file_name) as json_file:
            data = json.load(json_file)
        return data

    @keyword(name=None, tags=("Settings",))
    def load_payload(self, file_name):
        """*Load Request Payload as JSON file. Returns requestPayload as JSON Object*

        *Arguments*

            ``file_name``: json file name (with absolute path)

        *Examples*

        | ${requestPayload}=  |  `Load Payload`  | /path/to/file.json |
        """
        logger.debug("Check if file exists")
        if os.path.isfile(file_name) is False:
            logger.error("JSON file: " + file_name + " not found")
            raise IOError
        elif os.path.getsize(file_name) is 0:
            logger.error("JSON file: " + file_name + " is empty")
            raise IOError
        with open(file_name) as json_file:
            data = json.load(json_file)
        return data

    @keyword(name=None, tags=("Settings",))
    def update_header(self, request_headers, header_name, header_value):
        """*Update value of request header by providing header name and header value. Returns updatedHeaders as JSON Object*

        *Arguments*

            ``request_headers``: json object

            ``header_name``: Name of Request Header

            ``header_value``: Value of Request Header

        *Examples*

        | ${updatedHeaders}=  |  `Update Header` | ${requestHeaders} |  X-DNBAPI-UserId  |  TBxxxxx |
        """
        updated_headers = dict(request_headers)
        json_path_expr = parse('$' + '.' + header_name)
        matches = json_path_expr.find(updated_headers)
        if not matches:
            raise AssertionError("Header '%s' " % (header_name) +
                             "did not match anything.")
        for match in matches:
            path = match.path
            if isinstance(path, Index):
                match.context.value[match.path.index] = header_value
            elif isinstance(path, Fields):
                match.context.value[match.path.fields[0]] = header_value

        return updated_headers

    @keyword(name=None, tags=("Settings",))
    def add_headers(self, request_headers, new_headers):
        """*Add a dictionary or list object to existing json header object. Returns updatedHeaders as JSON Object*

    	*Arguments*

        	``request_headers``: json object

        	``new_headers``: dictionary to be added to request_headers

    	*Examples*

    	| ${newHeaders}=     |   Create Dictionary    | latitude=13.1234 	| longitude=130.1234       |
    	| ${updatedHeaders}= |  `Add Headers`         | ${requestHeaders}	| ${newHeaders}            |
    	"""
        updated_headers = dict(request_headers)
        json_path_expr = parse("$")
        for match in json_path_expr.find(updated_headers):
            if type(match.value) is dict:
                match.value.update(new_headers)
            if type(match.value) is list:
                match.value.append(new_headers)
        return updated_headers
        
    @keyword(name=None, tags=("Settings",))
    def add_header(self, request_headers, header_name, header_value):
        """*Add a dictionary or list object to existing json header object. Returns updatedHeaders as JSON Object*

    	*Arguments*

        	``request_headers``: json object

        	``header_name``: name of header to be included
        	
        	``header_value``: value of header to be included

    	*Examples*

    	| ${updatedHeaders}= |  `Add Header`         | ${requestHeaders}	| Authorization | Basic xxxx |
    	"""
        
        new_headers = { header_name : header_value}
        updated_headers = dict(request_headers)
        json_path = parse('$' + '.' + header_name)
        matches = json_path.find(updated_headers)
        if matches:
            raise AssertionError("Header '%s' " % (header_name) +
                             "exists.")
        if not matches:
            json_path_expr = parse("$")
            for match in json_path_expr.find(updated_headers):
                if type(match.value) is dict:
                    match.value.update(new_headers)
                if type(match.value) is list:
                    match.value.append(new_headers)
            return updated_headers
        

    @keyword(name=None, tags=("Settings",))
    def set_path_param(self, endpoint, paramName, paramValue):
        """*Replaces paramName in the given endpoint with paramValue. Returns updatedEndpoint*    

        *Arguments*                                                                      

            ``endpoint``: REST API Service URL with path parameter. For ex: https://testserver/v1/accounts/{accountNumber}/balance

            ``paramName``: Used as a parameterized string. For ex: {accountNumber}                     

            ``paramValue``: Used as a string that is replaced in the paramName provided. 	    	                                                  

        *Examples*                                                                       

        | ${updatedEndpoint} | `Set Path Param` | https://testserver/v1/accounts/{accountNumber}/balance | {accountNumber} | 12085214800 |    
        """
        return endpoint.replace(paramName, paramValue)

    @keyword(name=None, tags=("Settings",))
    def set_query_param(self, endpoint, paramName, paramValue):
        """*Sets the paramName & paramValue in the given endpoint. Returns updatedEndpoint*    

        *Arguments*                                                                      

            ``endpoint``: REST API Service URL. For ex: https://testserver/v1/accounts

            ``paramName``: Used as a Query Parameter string. For ex: accountNumber                     

            ``paramValue``: Used as a string that is added as value of Query Parameter. 	    	                                                  

        *Examples*                                                                       

        | ${updatedEndpoint} | `Set Query Param` | https://testserver/v1/accounts | accountNumber | 12085214800 |    
        """
        if "?" in endpoint:
            updatedEndpoint= endpoint + "&" + paramName + "=" + paramValue
            return updatedEndpoint
        
        if "?" not in endpoint:
            updatedEndpoint=endpoint + "?" + paramName + "=" + paramValue
            return updatedEndpoint

    
    @keyword(name=None, tags=("Assertions",))
    def get_value_from_response(self, response, json_path):
        """*Get Value From Response using JSONPath.  

        Arguments:
            - response: json object. 
            - json_path: jsonpath expression

        Return array of values

        Examples:
        | ${response}=  |  `Get` https://jsonplaceholder.typicode.com/users?_limit=1  |  
        | ${values}=  |  `Get Value From Response`  | ${response} |  $..phone_number |
        """
        json_path_expr = parse(json_path)
        match = json_path_expr.find(response)
        if not match:
            raise AssertionError("No match found with json path '%s' " % (json_path))    
        if match:
            return [match.value for match in json_path_expr.find(response)]
    
    
    ### Internal methods

    def _request(self, endpoint, request, validate=True):
        if not endpoint.startswith(('http://', 'https://')):
            base_url = self.request['scheme'] + "://" + self.request['netloc']
            if not endpoint.startswith('/'):
                endpoint = "/" + endpoint
            endpoint = urljoin(base_url, self.request['path']) + endpoint
        request['url'] = endpoint
        url_parts = urlparse(request['url'])
        request['scheme'] = url_parts.scheme
        request['netloc'] = url_parts.netloc
        request['path'] = url_parts.path
        try:
            response = client(request['method'], request['url'],
                              params=request['query'],
                              json=request['body'],
                              headers=request['headers'],
                              proxies=request['proxies'],
                              cert=request['cert'],
                              timeout=tuple(request['timeout']),
                              allow_redirects=request['allowRedirects'],
                              verify=request['sslVerify'])
        except SSLError as e:
            raise AssertionError("%s to %s SSL certificate verify failed:\n%s" %
                                 (request['method'], request['url'], e))
        except Timeout as e:
            raise AssertionError("%s to %s timed out:\n%s" % (
                request['method'], request['url'], e))
        utc_datetime = datetime.now(tz=utc)
        request['timestamp'] = {
            'utc': utc_datetime.isoformat(),
            'local': utc_datetime.astimezone(get_localzone()).isoformat()
        }
        if validate and self.spec:
            self._assert_spec(self.spec, response)
        instance = self._instantiate(request, response, validate)
        self.instances.append(instance)
        return instance

    def _instantiate(self, request, response, validate_schema=True):
        try:
            response_body = response.json()
        except ValueError:
            response_body = response.text
            if response_body:
                logger.warn("Response body content is not JSON. " +
                            "Content-Type is: %s" % response.headers['Content-Type'])
        response = {
            'seconds': response.elapsed.microseconds / 1000 / 1000,
            'status': response.status_code,
            'body': response_body,
            'headers': dict(response.headers)
        }
        schema = deepcopy(self.schema)
        schema['title'] = "%s %s" % (request['method'], request['url'])
        schema['description'] = "%s: %s" % (
            BuiltIn().get_variable_value("${SUITE NAME}"),
            BuiltIn().get_variable_value("${TEST NAME}")
        )
        request_properties = schema['properties']['request']['properties']
        response_properties = schema['properties']['response']['properties']
        if validate_schema:
            if request_properties:
                self._validate_schema(request_properties, request)
            if response_properties:
                self._validate_schema(response_properties, response)
        request_properties['body'] = self._new_schema(request['body'])
        request_properties['query'] = self._new_schema(request['query'])
        response_properties['body'] = self._new_schema(response['body'])
        if 'default' in schema and schema['default']:
            self._add_defaults_to_schema(schema, response)
        return {
            'request': request,
            'response': response,
            'schema': schema,
            'spec': self.spec
        }

    def _assert_spec(self, spec, response):
        request = response.request
        try:
            validate_api_call(spec, raw_request=request, raw_response=response)
        except ValueError as e:
            raise AssertionError(e)

    def _validate_schema(self, schema, json_dict):
        for field in schema:
            self._assert_schema(schema[field], json_dict[field])

    def _assert_schema(self, schema, reality):
        try:
            validate(reality, schema, format_checker=FormatChecker())
        except SchemaError as e:
            raise RuntimeError(e)
        except ValidationError as e:
            raise AssertionError(e)

    def _new_schema(self, value):
        builder = SchemaBuilder(schema_uri=False)
        builder.add_object(value)
        return builder.to_schema()

    def _add_defaults_to_schema(self, schema, response):
        body = response['body']
        schema = schema['properties']['response']['properties']['body']
        if isinstance(body, (dict)) and 'properties' in schema:
            self._add_property_defaults(body, schema['properties'])

    def _add_property_defaults(self, body, schema):
        for key in body:
            if "properties" in schema[key]:
                self._add_property_defaults(body[key], schema[key]['properties'])
            else:
                schema[key]['default'] = body[key]

    def _find_by_field(self, field, response=None, return_schema=True, print_found=True):
        last_instance = self._last_instance_or_error(response)
        schema = None
        paths = []
        if field.startswith("$"):
            value = last_instance['response']['body']
            if return_schema:
                schema = last_instance['schema']['properties']['response']
                schema = schema['properties']['body']
            if field == "$":
                return [{
                    'path': ["response", "body"],
                    'reality': value,
                    'schema': schema
                }]
            try:
                query = parse_jsonpath(field)
            except Exception as e:
                raise RuntimeError("Invalid JSONPath query '%s': %s" % (
                    field, e))
            matches = [str(match.full_path) for match in query.find(value)]
            if not matches:
                raise AssertionError("JSONPath query '%s' " % (field) +
                    "did not match anything.")
            for match in matches:
                path = match.replace("[", "").replace("]", "").split('.')
                paths.append(path)
        else:
            value = last_instance
            if return_schema:
                schema = last_instance['schema']['properties']
            path = field.split()
            paths.append(path)
        return [self._find_by_path(field, path, value, schema, print_found)
                for path in paths]

    def _last_instance_or_error(self, response=None):
        if response:
            return self.input(response)
        else:
            try:
                return self.instances[-1]
            except IndexError:
                raise RuntimeError("No instances: No requests made, " +
                    "and no previous instances loaded in the library import.")

    def _find_by_path(self, field, path, value, schema=None, print_found=True):
        for key in path:
            try:
                value = self._value_by_key(value, key)
            except (KeyError, TypeError):
                if print_found:
                    self.log_json(value,
                        "\n\nProperty '%s' does not exist in:" % (key))
                raise AssertionError(
                    "\nExpected property '%s' was not found." % (field))
            except IndexError:
                if print_found:
                    self.log_json(value,
                        "\n\nIndex '%s' does not exist in:" % (key))
                raise AssertionError(
                    "\nExpected index '%s' did not exist." % (field))
            if schema:
                schema = self._schema_by_key(schema, key, value)
        found = {
            'path': path,
            'reality': value,
            'schema': schema
        }
        return found

    def _value_by_key(self, json, key):
        try:
            return json[int(key)]
        except ValueError:
            return json[key]

    def _schema_by_key(self, schema, key, value):
        if 'properties' in schema:
            schema = schema['properties']
        elif 'items' in schema:
            if isinstance(schema['items'], (dict)):
                schema['items'] = [schema['items']]
            new_schema = self._new_schema(value)
            try:
                return schema['items'][schema['items'].index(new_schema)]
            except ValueError:
                schema['items'].append(new_schema)
                return schema['items'][-1]
        if key not in schema:
            schema[key] = self._new_schema(value)
        return schema[key]

    def _should_add_examples(self):
        return 'examples' in self.schema and isinstance(
            self.schema['examples'], (list))

    def _set_type_validations(self, json_type, schema, validations):
        if validations:
            if "draft-04" in self.schema['$schema']:
                schema_version = "draft-04"
            elif "draft-06" in self.schema['$schema']:
                schema_version = "draft-06"
            else:
                schema_version = "draft-07"
            kws = list(SCHEMA_KEYWORDS['common'][schema_version])
            kws.extend(SCHEMA_KEYWORDS[json_type][schema_version])
        for validation in validations:
            if validation not in kws:
                raise RuntimeError("Unknown JSON Schema (%s)" % (
                    schema_version) + " validation keyword " +
                                   "for %s:\n%s" % (json_type, validation))
            schema[validation] = self.input(validations[validation])
        schema.update({"type": json_type})



# -*- encoding: utf-8 -*-

#  Copyright 2019-  DNB
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from rbfZOS.x3270 import x3270



__version__ = "1.0"


class rbfZOS(x3270):
    """rbfZOS is a library for Robot Framework based on [https://pypi.org/project/py3270/|py3270 project],
       a Python interface to x3270, an IBM 3270 terminal emulator. It provides an API to a x3270 or s3270 subprocess.
       = Installation  =
       
       For use this library you need to install the [http://x3270.bgp.nu/download.html|x3270 project]
       and put the directory on your PATH. On Windows, you need to download wc3270 and put
       the "C:\Program Files\wc3270" in PATH of the Environment Variables.
              
       = Notes  =
       By default the import set the visible argument to true, on this option the py3270 is running the wc3270.exe,
       but is you set the visible to false, the py3270 will run the ws3270.exe.
             
       
       
       = Mainframe Example =
       | ***** Settings *****
       | Library           rbfZOS
       | Library           BuiltIn
       |
       | ***** Test Cases *****
       | Example
       |     Open Connection    Hostname    LUname
       |     Change Wait Time    0.9
       |     Set Screenshot Folder    C:\\\Temp\\\IMG
       |     ${value}    Read    3    10    17
       |     Page Should Contain String    ENTER APPLICATION
       |     Write Bare    applicationname
       |     Send Enter
       |     Take Screenshot
       |     Close Connection 
    """
    ROBOT_LIBRARY_SCOPE = 'GLOBAL'
    ROBOT_LIBRARY_VERSION = __version__



from .application import ApplicationKeywords  # noqa: F401 #pylint: disable=unused-import
from .keyboard import KeyboardKeywords  # noqa: F401 #pylint: disable=unused-import
from .window import WindowKeywords  # noqa: F401 #pylint: disable=unused-import
from .screenshot import ScreenshotKeywords  # noqa: F401 #pylint: disable=unused-import
from .configuration import WhiteConfigurationKeywords  # noqa: F401 #pylint: disable=unused-import



# Copyright 2019-     DNB

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .alert import AlertKeywords
from .browsermanagement import BrowserManagementKeywords
from .cookie import CookieKeywords
from .element import ElementKeywords
from .formelement import FormElementKeywords
from .frames import FrameKeywords
from .javascript import JavaScriptKeywords
from .runonfailure import RunOnFailureKeywords
from .screenshot import ScreenshotKeywords
from .selectelement import SelectElementKeywords
from .tableelement import TableElementKeywords
from .waiting import WaitingKeywords
from .webdrivertools import WebDriverCache
from .webdrivertools import WebDriverCreator
from .window import WindowKeywords
from .nettbank import NettbankKeywords



# -*- coding: utf-8 -*-

#  Copyright 2019-  DNB
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


import csv
import os


class PictReader:

    def __init__(self, file, testcase_table_name):
        self.file = file
        self.filename, self.file_extension = os.path.splitext(self.file)
        self.testcase_table_name = testcase_table_name
        self.data_table = None

    def get_data_from_pict(self):
        self._register_dialect()
        self._create_gemerated_file_from_model_file()
        self._read_generated_file_to_dictionaries()
        return self.data_table

    def _register_dialect(self):
        csv.register_dialect('PICT',
                             delimiter='\t',
                             skipinitialspace=False,
                             lineterminator='\r\n',
                             quoting=csv.QUOTE_NONE)

    def _create_gemerated_file_from_model_file(self):
        os.system(f'pict "{self.file}" > "{self.filename}.pictout"')

    def _read_generated_file_to_dictionaries(self):
        with open(f'{self.filename}.pictout', 'r', encoding='utf_8') as csvfile:
            reader = csv.reader(csvfile, 'PICT')
            table = {}
            header = []
            test_case_column = []
            for row_index, row in enumerate(reader):
                if row_index == 0:
                    header = row
                    for header_cell in header:
                        table[f'${{{header_cell}}}'] = []
                else:
                    for cell_index, cell in enumerate(row):
                        table[f'${{{header[cell_index]}}}'].append(cell)
                    test_case_column.append('')
            table[self.testcase_table_name] = test_case_column
        self.data_table = table



# Copyright 2019-     DNB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from rbfWEB.base import LibraryComponent, keyword
from rbfWEB.utils import is_noney, is_string


class RunOnFailureKeywords(LibraryComponent):

    @keyword
    def register_keyword_to_run_on_failure(self, keyword):
        """Sets the keyword to execute when a rbfWEB keyword fails.

        ``keyword`` is the name of a keyword that will be executed if a
        rbfWEB keyword fails. It is possible to use any available
        keyword, including user keywords or keywords from other libraries,
        but the keyword must not take any arguments.

        The initial keyword to use is set when `importing` the library, and
        the keyword that is used by default is `Capture Page Screenshot`.
        Taking a screenshot when something failed is a very useful
        feature, but notice that it can slow down the execution.

        It is possible to use string ``NOTHING`` or ``NONE``,
        case-insensitively, as well as Python ``None`` to disable this
        feature altogether.

        This keyword returns the name of the previously registered
        failure keyword or Python ``None`` if this functionality was
        previously disabled. The return value can be always used to
        restore the original value later.

        Example:
        | `Register Keyword To Run On Failure`  | Log Source |
        | ${previous kw}= | `Register Keyword To Run On Failure`  | NONE |
        | `Register Keyword To Run On Failure`  | ${previous kw} |

        
        """
        old_keyword = self.ctx.run_on_failure_keyword
        keyword = self.resolve_keyword(keyword)
        self.ctx.run_on_failure_keyword = keyword
        self.info('%s will be run on failure.' % (keyword or 'No keyword'))
        return old_keyword

    @staticmethod
    def resolve_keyword(name):
        if is_noney(name) or is_string(name) and name.upper() == 'NOTHING':
            return None
        return name



# -*- coding: utf-8 -*-

#  Copyright 2019-  DNB
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import importlib
from robot.api import logger
from robot.api.deco import keyword


try:
    import ConfigParser
except:
    import configparser as ConfigParser

from robot.api import logger


class ConnectionManager(object):
    """
    Connection Manager handles the connection & disconnection to the database.
    """

    def __init__(self):
        """
        Initializes _dbconnection to None.
        """
        self._dbconnection = None
        self.db_api_module_name = None

    
    @keyword(name=None, tags=("Connection",))
    def connect_to_database(self, dbapiModuleName=None, dbName=None, dbUsername=None, dbPassword=None, dbHost=None, dbPort=None, dbCharset=None, dbConfigFile="./resources/db.cfg"):
        """
        *Loads the DB API 2.0 module given `dbapiModuleName` then uses it to
        connect to the database using `dbName`, `dbUsername`, and `dbPassword`.*
        
        *Arguments*

        ``dbapiModuleName``: Type of database. For ex: ibm_db, ibm_db_dbi, MySQLdb, pymysql, pyodbc, psycopg2, etc.,
        
        ``dbName``: Name of database.
        
        ``dbUsername``: Username  
        
        ``dbPassword``: Password 
        
        ``dbHost``: Database server name
        
        ``dbPort``: Database port
        
        ``dbCharset``: Charset
        
        ``dbConfigFile``: Optionally, you can specify a `dbConfigFile` wherein it will load the default property values for `dbapiModuleName`, `dbName` `dbUsername` and `dbPassword` 
        (note: specifying `dbapiModuleName`, `dbName`, `dbUsername` or `dbPassword` directly will override the properties of the same key in `dbConfigFile`). 
        If no `dbConfigFile` is specified, it defaults to `./resources/db.cfg`.

        The `dbConfigFile` is useful if you don't want to check into your SCM your database credentials.

        Example db.cfg file
        | [default]
        | dbapiModuleName=pymysqlforexample
        | dbName=yourdbname
        | dbUsername=yourusername
        | dbPassword=yourpassword
        | dbHost=yourhost
        | dbPort=yourport    

        *Examples*
              
        | # explicitly specifies all db property values |
        | `Connect To Database` | psycopg2 | my_db | postgres | s3cr3t | tiger.foobar.com | 5432 |

        | # loads all property values from default.cfg |
        | `Connect To Database` | dbConfigFile=default.cfg |

        | # loads all property values from ./resources/db.cfg |
        | `Connect To Database` |

        | # uses explicit `dbapiModuleName` and `dbName` but uses the `dbUsername` and `dbPassword` in 'default.cfg' |
        | `Connect To Database` | psycopg2 | my_db_test | dbConfigFile=default.cfg |

        | # uses explicit `dbapiModuleName` and `dbName` but uses the `dbUsername` and `dbPassword` in './resources/db.cfg' |
        | `Connect To Database` | psycopg2 | my_db_test |
        """

        config = ConfigParser.ConfigParser()
        config.read([dbConfigFile])

        dbapiModuleName = dbapiModuleName or config.get('default', 'dbapiModuleName')
        dbName = dbName or config.get('default', 'dbName')
        dbUsername = dbUsername or config.get('default', 'dbUsername')
        dbPassword = dbPassword if dbPassword is not None else config.get('default', 'dbPassword')
        dbHost = dbHost or config.get('default', 'dbHost') or 'localhost'
        dbPort = int(dbPort or config.get('default', 'dbPort'))

        if dbapiModuleName == "excel" or dbapiModuleName == "excelrw":
            self.db_api_module_name = "pyodbc"
            db_api_2 = importlib.import_module("pyodbc")
        else:
            self.db_api_module_name = dbapiModuleName
            db_api_2 = importlib.import_module(dbapiModuleName)
        if dbapiModuleName in ["MySQLdb", "pymysql"]:
            dbPort = dbPort or 3306
            logger.info('Connecting using : %s.connect(db=%s, user=%s, passwd=%s, host=%s, port=%s, charset=%s) ' % (dbapiModuleName, dbName, dbUsername, dbPassword, dbHost, dbPort, dbCharset))
            self._dbconnection = db_api_2.connect(db=dbName, user=dbUsername, passwd=dbPassword, host=dbHost, port=dbPort, charset=dbCharset)
        elif dbapiModuleName in ["psycopg2"]:
            dbPort = dbPort or 5432
            logger.info('Connecting using : %s.connect(database=%s, user=%s, password=%s, host=%s, port=%s) ' % (dbapiModuleName, dbName, dbUsername, dbPassword, dbHost, dbPort))
            self._dbconnection = db_api_2.connect(database=dbName, user=dbUsername, password=dbPassword, host=dbHost, port=dbPort)
        elif dbapiModuleName in ["pyodbc", "pypyodbc"]:
            dbPort = dbPort or 1433
            logger.info('Connecting using : %s.connect(DRIVER={SQL Server};SERVER=%s,%s;DATABASE=%s;UID=%s;PWD=%s)' % (dbapiModuleName, dbHost, dbPort, dbName, dbUsername, dbPassword))
            self._dbconnection = db_api_2.connect('DRIVER={SQL Server};SERVER=%s,%s;DATABASE=%s;UID=%s;PWD=%s' % (dbHost, dbPort, dbName, dbUsername, dbPassword))
        elif dbapiModuleName in ["excel"]:
            logger.info(
                'Connecting using : %s.connect(DRIVER={Microsoft Excel Driver (*.xls, *.xlsx, *.xlsm, *.xlsb)};DBQ=%s;ReadOnly=1;Extended Properties="Excel 8.0;HDR=YES";)' % (
                dbapiModuleName, dbName))
            self._dbconnection = db_api_2.connect(
                'DRIVER={Microsoft Excel Driver (*.xls, *.xlsx, *.xlsm, *.xlsb)};DBQ=%s;ReadOnly=1;Extended Properties="Excel 8.0;HDR=YES";)' % (
                    dbName), autocommit=True)
        elif dbapiModuleName in ["excelrw"]:
            logger.info(
                'Connecting using : %s.connect(DRIVER={Microsoft Excel Driver (*.xls, *.xlsx, *.xlsm, *.xlsb)};DBQ=%s;ReadOnly=0;Extended Properties="Excel 8.0;HDR=YES";)' % (
                dbapiModuleName, dbName))
            self._dbconnection = db_api_2.connect(
                'DRIVER={Microsoft Excel Driver (*.xls, *.xlsx, *.xlsm, *.xlsb)};DBQ=%s;ReadOnly=0;Extended Properties="Excel 8.0;HDR=YES";)' % (
                    dbName), autocommit=True)
        elif dbapiModuleName in ["ibm_db", "ibm_db_dbi"]:
            dbPort = dbPort or 50000
            logger.info('Connecting using : %s.connect(DATABASE=%s;HOSTNAME=%s;PORT=%s;PROTOCOL=TCPIP;UID=%s;PWD=%s;) ' % (dbapiModuleName, dbName, dbHost, dbPort, dbUsername, dbPassword))
            self._dbconnection = db_api_2.connect('DATABASE=%s;HOSTNAME=%s;PORT=%s;PROTOCOL=TCPIP;UID=%s;PWD=%s;' % (dbName, dbHost, dbPort, dbUsername, dbPassword), '', '')
        elif dbapiModuleName in ["cx_Oracle"]:
            dbPort = dbPort or 1521
            oracle_dsn =  db_api_2.makedsn(host=dbHost, port=dbPort, service_name=dbName)
            logger.info('Connecting using: %s.connect(user=%s, password=%s, dsn=%s) ' % (dbapiModuleName, dbUsername, dbPassword, oracle_dsn))
            self._dbconnection = db_api_2.connect(user=dbUsername, password=dbPassword, dsn=oracle_dsn)
        else:
            logger.info('Connecting using : %s.connect(database=%s, user=%s, password=%s, host=%s, port=%s) ' % (dbapiModuleName, dbName, dbUsername, dbPassword, dbHost, dbPort))
            self._dbconnection = db_api_2.connect(database=dbName, user=dbUsername, password=dbPassword, host=dbHost, port=dbPort)

    
    @keyword(name=None, tags=("Connection",))
    def connect_to_database_using_custom_params(self, dbapiModuleName=None, db_connect_string=''):
        """
        *Loads the DB API 2.0 module given `dbapiModuleName` then uses it to connect to the database using the map string `db_custom_param_string`.*
        
        *Arguments*

        ``dbapiModuleName``: Type of database. For ex: ibm_db, ibm_db_dbi, MySQLdb, pymysql, pyodbc, psycopg2, etc.,
        
        ``db_connect_string``: Database connection paramaters
        
        *Examples*

        | # for psycopg2 |
        | `Connect To Database Using Custom Params` | psycopg2 | database='my_db_test', user='postgres', password='s3cr3t', host='tiger.foobar.com', port=5432 |

        | # for JayDeBeApi |
        | `Connect To Database Using Custom Params` | JayDeBeApi | 'oracle.jdbc.driver.OracleDriver', 'my_db_test', 'system', 's3cr3t' |
        """
        db_api_2 = importlib.import_module(dbapiModuleName)

        db_connect_string = 'db_api_2.connect(%s)' % db_connect_string

        self.db_api_module_name = dbapiModuleName
        logger.info('Executing : Connect To Database Using Custom Params : %s.connect(%s) ' % (dbapiModuleName, db_connect_string))
        self._dbconnection = eval(db_connect_string)

    
    @keyword(name=None, tags=("Connection",))
    def disconnect_from_database(self):
        """
        *Disconnects from the database.*

        *Examples*
        
        | `Disconnect From Database` | # disconnects from current connection to the database |
        """
        logger.info('Executing : Disconnect From Database')
        self._dbconnection.close()

    
    @keyword(name=None, tags=("Connection",))
    def set_auto_commit(self, autoCommit=True):
        """
        *Turn the autocommit on the database connection ON or OFF.* 
        
        The default behaviour on a newly created database connection is to automatically start a 
        transaction, which means that database actions that won't work if there is an active 
        transaction will fail. Common examples of these actions are creating or deleting a database 
        or database snapshot. By turning on auto commit on the database connection these actions 
        can be performed.
        
        *Arguments*

        ``autoCommit``: Defaults to TRUE. Desired state can be explicitly set. 
        
        *Examples*
        
        | # Explicitly set the desired state |
        | `Set Auto Commit` | False |
        """
        logger.info('Executing : Set Auto Commit')
        self._dbconnection.autocommit = autoCommit



# Copyright 2019-     DNB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from rflint.common import TestRule, ERROR, WARNING
from rflint.parser import SettingTable


class PeriodInTestName(TestRule):
    '''Warn about periods in the testcase name
    
    Since robot uses "." as a path separator, using a "." in a testcase
    name can lead to ambiguity. 
    '''
    severity = WARNING
    
    def apply(self,testcase):
        if "." in testcase.name:
            self.report(testcase, "'.' in testcase name '%s'" % testcase.name, testcase.linenumber)

class TagWithSpaces(TestRule):
    '''Flags tags that have spaces in the tag name'''
    severity=ERROR

    def apply(self, testcase):
        for tag in testcase.tags:
            if ((" " in tag) or ("\t" in tag)):
                self.report(testcase, "space not allowed in tag name: '%s'" % tag, testcase.linenumber)

class RequireTestDocumentation(TestRule):
    '''Verify that a test suite has documentation

    This rule is not enforced for data driven tests ("Test Template" in Settings)
    '''
    severity=ERROR

    def apply(self, testcase):
        if testcase.is_templated:
            return

        for setting in testcase.settings:
            if setting[1].lower() == "[documentation]" and len(setting) > 2:
                return

        # set the line number to the line immediately after the testcase name
        self.report(testcase, "No testcase documentation", testcase.linenumber+1)

#class TooFewTestSteps(TestRule):
#    '''Tests should have at least a minimum number of steps

#    This rule is configurable. The default number of required steps is 2.
#    '''

#    min_required = 2

#    def configure(self, min_required):
#        self.min_required = int(min_required)

#    def apply(self, testcase):
#        if testcase.is_templated:
#            return

        # ignore empty test steps
#        steps = [step for step in testcase.steps if not (len(step) == 1
                                                         and not step[0])]
#        if len(steps) < self.min_required:
#            msg = "Too few steps (%s) in test case" % len(steps)
#            self.report(testcase, msg,  testcase.linenumber)

#class TooManyTestSteps(TestRule):
#    '''Workflow tests should have no more than ten steps.

#    https://code.google.com/p/robotframework/wiki/HowToWriteGoodTestCases#Workflow_tests
#    '''

#    severity=WARNING
#    max_allowed = 10

#    def configure(self, max_allowed):
#        self.max_allowed = int(max_allowed)

#    def apply(self, testcase):
#        if testcase.is_templated:
#            return

        # ignore empty test steps
#        steps = [step for step in testcase.steps if not (len(step) == 1
#                                                         and not step[0])]
#        if len(steps) > self.max_allowed:
#            self.report(testcase,
#                        "Too many steps (%s) in test case" % len(steps),
#                        steps[self.max_allowed].startline)



# Copyright 2019-     DNB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time

from selenium.common.exceptions import StaleElementReferenceException

from rbfWEB.base import LibraryComponent, keyword
from rbfWEB.errors import ElementNotFound
from rbfWEB.utils import is_noney, secs_to_timestr


class WaitingKeywords(LibraryComponent):

    @keyword
    def wait_for_condition(self, condition, timeout=None, error=None):
        """Waits until ``condition`` is true or ``timeout`` expires.

        The condition can be arbitrary JavaScript expression but it
        must return a value to be evaluated. See `Execute JavaScript` for
        information about accessing content on pages.

        Fails if the timeout expires before the condition becomes true. See
        the `Timeouts` section for more information about using timeouts
        and their default value.

        ``error`` can be used to override the default error message.

        Examples:
        | `Wait For Condition` | return document.title == "New Title" |
        | `Wait For Condition` | return jQuery.active == 0            |
        | `Wait For Condition` | style = document.querySelector('h1').style; return style.background == "red" && style.color == "white" |
        """
        if 'return' not in condition:
            raise ValueError("Condition '%s' did not have mandatory 'return'."
                             % condition)
        self._wait_until(
            lambda: self.driver.execute_script(condition) is True,
            "Condition '%s' did not become true in <TIMEOUT>." % condition,
            timeout, error
        )

    @keyword
    def wait_until_location_is(self, expected, timeout=None, message=None):
        """Wait until that current URL is ``expected``.

        The ``expected`` argument is the expected value in url.

        Fails if ``timeout`` expires before the location is. See
        the `Timeouts` section for more information about using timeouts
        and their default value.

        The ``message`` argument can be used to override the default error
        message.

        
        """

        expected = str(expected)
        self._wait_until(lambda: expected == self.driver.current_url,
                         "Location did not is '%s' in <TIMEOUT>." % expected,
                         timeout, message)

    @keyword
    def wait_until_location_contains(self, expected, timeout=None, message=None):
        """Wait until that current URL contains ``expected``.

        The ``expected`` argument contains the expected value in url.

        Fails if ``timeout`` expires before the location contains. See
        the `Timeouts` section for more information about using timeouts
        and their default value.

        The ``message`` argument can be used to override the default error
        message.

        
        """
        expected = str(expected)
        self._wait_until(lambda: expected in self.driver.current_url,
                         "Location did not contain '%s' in <TIMEOUT>." % expected,
                         timeout, message)


    @keyword
    def wait_until_page_contains(self, text, timeout=None, error=None):
        """Waits until ``text`` appears on current page.

        Fails if ``timeout`` expires before the text appears. See
        the `Timeouts` section for more information about using timeouts
        and their default value.

        ``error`` can be used to override the default error message.
        """
        self._wait_until(lambda: self.is_text_present(text),
                         "Text '%s' did not appear in <TIMEOUT>." % text,
                         timeout, error)

    @keyword
    def wait_until_page_does_not_contain(self, text, timeout=None,
                                         error=None):
        """Waits until ``text`` disappears from current page.

        Fails if ``timeout`` expires before the text disappears. See
        the `Timeouts` section for more information about using timeouts
        and their default value.

        ``error`` can be used to override the default error message.
        """
        self._wait_until(lambda: not self.is_text_present(text),
                         "Text '%s' did not disappear in <TIMEOUT>." % text,
                         timeout, error)

    @keyword
    def wait_until_page_contains_element(self, locator, timeout=None,
                                         error=None):
        """Waits until element ``locator`` appears on current page.

        Fails if ``timeout`` expires before the element appears. See
        the `Timeouts` section for more information about using timeouts and
        their default value and the `Locating elements` section for details
        about the locator syntax.

        ``error`` can be used to override the default error message.
        """
        self._wait_until(
            lambda: self.find_element(locator, required=False) is not None,
            "Element '%s' did not appear in <TIMEOUT>." % locator,
            timeout, error
        )

    @keyword
    def wait_until_page_does_not_contain_element(self, locator, timeout=None,
                                                 error=None):
        """Waits until element ``locator`` disappears from current page.

        Fails if ``timeout`` expires before the element disappears. See
        the `Timeouts` section for more information about using timeouts and
        their default value and the `Locating elements` section for details
        about the locator syntax.

        ``error`` can be used to override the default error message.
        """
        self._wait_until(
            lambda: self.find_element(locator, required=False) is None,
            "Element '%s' did not disappear in <TIMEOUT>." % locator,
            timeout, error
        )

    @keyword
    def wait_until_element_is_visible(self, locator, timeout=None,
                                      error=None):
        """Waits until element ``locator`` is visible.

        Fails if ``timeout`` expires before the element is visible. See
        the `Timeouts` section for more information about using timeouts and
        their default value and the `Locating elements` section for details
        about the locator syntax.

        ``error`` can be used to override the default error message.
        """
        self._wait_until(
            lambda: self.is_visible(locator),
            "Element '%s' not visible after <TIMEOUT>." % locator,
            timeout, error
        )

    @keyword
    def wait_until_element_is_not_visible(self, locator, timeout=None,
                                          error=None):
        """Waits until element ``locator`` is not visible.

        Fails if ``timeout`` expires before the element is not visible. See
        the `Timeouts` section for more information about using timeouts and
        their default value and the `Locating elements` section for details
        about the locator syntax.

        ``error`` can be used to override the default error message.
        """
        self._wait_until(
            lambda: not self.is_visible(locator),
            "Element '%s' still visible after <TIMEOUT>." % locator,
            timeout, error
        )

    @keyword
    def wait_until_element_is_enabled(self, locator, timeout=None,
                                      error=None):
        """Waits until element ``locator`` is enabled.

        Element is considered enabled if it is not disabled nor read-only.

        Fails if ``timeout`` expires before the element is enabled. See
        the `Timeouts` section for more information about using timeouts and
        their default value and the `Locating elements` section for details
        about the locator syntax.

        ``error`` can be used to override the default error message.

        
        """
        self._wait_until(
            lambda: self.is_element_enabled(locator),
            "Element '%s' was not enabled in <TIMEOUT>." % locator,
            timeout, error
        )

    @keyword
    def wait_until_element_contains(self, locator, text, timeout=None,
                                    error=None):
        """Waits until element ``locator`` contains ``text``.

        Fails if ``timeout`` expires before the text appears. See
        the `Timeouts` section for more information about using timeouts and
        their default value and the `Locating elements` section for details
        about the locator syntax.

        ``error`` can be used to override the default error message.
        """
        self._wait_until(
            lambda: text in self.find_element(locator).text,
            "Element '%s' did not get text '%s' in <TIMEOUT>." % (locator, text),
            timeout, error
        )

    @keyword
    def wait_until_element_does_not_contain(self, locator, text, timeout=None,
                                            error=None):
        """Waits until element ``locator`` does not contain ``text``.

        Fails if ``timeout`` expires before the text disappears. See
        the `Timeouts` section for more information about using timeouts and
        their default value and the `Locating elements` section for details
        about the locator syntax.

        ``error`` can be used to override the default error message.
        """
        self._wait_until(
            lambda: text not in self.find_element(locator).text,
            "Element '%s' still had text '%s' after <TIMEOUT>." % (locator, text),
            timeout, error
        )

    def _wait_until(self, condition, error, timeout=None, custom_error=None):
        timeout = self.get_timeout(timeout)
        if is_noney(custom_error):
            error = error.replace('<TIMEOUT>', secs_to_timestr(timeout))
        else:
            error = custom_error
        self._wait_until_worker(condition, timeout, error)

    def _wait_until_worker(self, condition, timeout, error):
        max_time = time.time() + timeout
        not_found = None
        while time.time() < max_time:
            try:
                if condition():
                    return
            except ElementNotFound as err:
                not_found = str(err)
            except StaleElementReferenceException as err:
                self.info('Suppressing StaleElementReferenceException from Selenium.')
                not_found = err
            else:
                not_found = None
            time.sleep(0.2)
        raise AssertionError(not_found or error)



from robot.api import logger  # noqa:  F401 #pylint: disable=unused-import
from TestStack.White.UIItems.ListBoxItems import ComboBox, ListBox
from TestStack.White.UIItems import UIActionException
from rbfDesktop.keywords.librarycomponent import LibraryComponent
from rbfDesktop.keywords.robotlibcore import keyword


class ListKeywords(LibraryComponent):
    @keyword
    def select_listbox_value(self, locator, value):
        """Selects a value from a listbox.

        ``locator`` is the locator of the listbox or ListBox item object.
        Locator syntax is explained in `Item locators`.

        ``value`` is the value to be selected.
        """
        listbox = self.state._get_typed_item_by_locator(ListBox, locator)
        listbox.Select(value)

    @keyword
    def select_listbox_index(self, locator, item_index):
        """Selects an item by its index from a listbox.

        ``locator`` is the locator of the listbox or ListBox item object.
        Locator syntax is explained in `Item locators`.

        ``item_index`` is the index of the item to select.
        """
        listbox = self.state._get_typed_item_by_locator(ListBox, locator)
        listbox.Select(int(item_index))

    @keyword
    def listbox_selection_should_be(self, locator, expected):
        """Checks the listbox selection.

        Fails if the selection was not as expected.

        ``locator`` is the locator of the listbox or ListBox item object.
        Locator syntax is explained in `Item locators`.

        ``expected`` is the expected selection value.
        """
        listbox = self.state._get_typed_item_by_locator(ListBox, locator)
        if listbox.SelectedItemText != expected:
            raise AssertionError(
                "Expected listbox selection to be {}, was {}".format(expected, listbox.SelectedItemText)
            )

    @keyword
    def listbox_should_contain(self, locator, expected):
        """Checks that listbox contains an item with text ``expected``.

        Fails if the listbox does not contain an item with the given text.

        ``locator`` is the locator of the listbox or ListBox item object.
        Locator syntax is explained in `Item locators`.

        ``expected`` is the expected item text.
        """
        listbox = self.state._get_typed_item_by_locator(ListBox, locator)
        try:
            listbox.Item(str(expected))
        except UIActionException as error:
            # Check error in case we get UIActionException with another message
            if "Item of text" in str(error):
                raise AssertionError("ListBox with locator '{}' did not contain '{}'".format(locator, expected))
            raise error

    @keyword
    def listbox_should_not_contain(self, locator, expected):
        """Verifies that a listbox does not contain an item with text ``expected``.

        Fails if the listbox contains an item with the given text.

        ``locator`` is the locator of the listbox or ListBox item object.
        Locator syntax is explained in `Item locators`.

        ``expected`` is the expected item text.
        """
        listbox = self.state._get_typed_item_by_locator(ListBox, locator)
        try:
            listbox.Item(str(expected))
            raise AssertionError("ListBox with locator '{}' should not have contained '{}'".format(locator, expected))
        except UIActionException as error:
            # Check error in case we get UIActionException with another message
            if "Item of text" not in str(error):
                raise error

    @keyword
    def select_combobox_value(self, locator, value):
        """Selects a value from a combobox.

        ``locator`` is the locator of the combobox or ComboBox item object.
        Locator syntax is explained in `Item locators`.

        ``value`` is the value to be selected.
        """
        combobox = self.state._get_typed_item_by_locator(ComboBox, locator)
        combobox.Select(value)

    @keyword
    def select_combobox_index(self, locator, item_index):
        """
        Selects a value from combobox by using its index.

        ``locator`` is the locator of the combobox or ComboBox item object.
        Locator syntax is explained in `Item locators`.

        ``item_index`` is the index to be selected.
        """
        combobox = self.state._get_typed_item_by_locator(ComboBox, locator)
        combobox.Select(int(item_index))

    @keyword
    def verify_combobox_item(self, locator, expected):
        """*DEPRECATED* Please use Verify Combobox Selection instead.

        Verifies the selected value of a combobox.

        ``locator`` is the locator of the combobox or ComboBox item object.
        Locator syntax is explained in `Item locators`.

        ``expected`` is the expected selection value.
        """
        self.verify_combobox_selection(locator, expected)

    @keyword
    def verify_combobox_selection(self, locator, expected):
        """Verifies that the combobox value is selected.

        ``locator`` is the locator of the combobox or ComboBox item object.
        Locator syntax is explained in `Item locators`.

        ``expected`` is the expected selection value.
        """
        combobox = self.state._get_typed_item_by_locator(ComboBox, locator)
        self.state._verify_value(expected, combobox.EditableText)

    @keyword
    def combobox_should_contain(self, locator, expected):
        """Verifies that a combobox contains an item with text ``expected``.

        Fails if the combobox does not contain an item with the given text.

        ``locator`` is the locator of the combobox or ComboBox item object.
        Locator syntax is explained in `Item locators`.

        ``expected`` is the expected item text.
        """
        combobox = self.state._get_typed_item_by_locator(ComboBox, locator)
        try:
            combobox.Item(str(expected))
        except UIActionException as error:
            # Check error in case we get UIActionException with another message
            if "Item of text" in str(error):
                raise AssertionError("ComboBox with locator '{}' did not contain '{}'".format(locator, expected))
            raise error

    @keyword
    def combobox_should_not_contain(self, locator, expected):
        """Verifies that a combobox does not contain an item with text ``expected``.

        Fails if the combobox contains an item with the given text.

        ``locator`` is the locator of the combobox or ComboBox item object.
        Locator syntax is explained in `Item locators`.

        ``expected`` is the expected item text.
        """
        combobox = self.state._get_typed_item_by_locator(ComboBox, locator)
        try:
            combobox.Item(str(expected))
            raise AssertionError("ComboBox with locator '{}' should not have contained '{}'".format(locator, expected))
        except UIActionException as error:
            # Check error in case we get UIActionException with another message
            if "Item of text" not in str(error):
                raise error



from System.Windows import Point
from TestStack.White.UIA import RectX
from TestStack.White.InputDevices import Mouse


class Clicks:
    # Low level function to handle offset click.
    @staticmethod
    def click(item, x_offset, y_offset):
        if x_offset == 0 and y_offset == 0:
            item.Click()
        else:
            offset_position = Clicks._get_offset_point(item, x_offset, y_offset)
            Mouse.Instance.Click(offset_position)

    # Low level helper function to handle offset right click.
    @staticmethod
    def right_click(item, x_offset, y_offset):
        if x_offset == 0 and y_offset == 0:
            item.RightClick()
        else:
            offset_position = Clicks._get_offset_point(item, x_offset, y_offset)
            Mouse.Instance.Location = offset_position
            Mouse.Instance.RightClick()

    # Low level helper function to handle offset double click.
    @staticmethod
    def double_click(item, x_offset, y_offset):
        if x_offset == 0 and y_offset == 0:
            item.DoubleClick()
        else:
            offset_position = Clicks._get_offset_point(item, x_offset, y_offset)
            Mouse.Instance.DoubleClick(offset_position)

    # Helper function to translate item center to offset point
    @staticmethod
    def _get_offset_point(item, x_offset, y_offset):
        item_bounds = item.Bounds
        item_center = RectX.Center(item_bounds)
        offset_point = Point(int(item_center.X) + int(x_offset), int(item_center.Y) + int(y_offset))
        if not item_bounds.Contains(offset_point):
            raise AssertionError("click location out of bounds")
        return offset_point



# Copyright 2019-     DNB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from robot.api import logger
from robot.utils import NormalizedDict
from selenium.webdriver.remote.webelement import WebElement

from rbfWEB.base import ContextAware
from rbfWEB.errors import ElementNotFound
from rbfWEB.utils import escape_xpath_value, events, is_falsy

from .customlocator import CustomLocator


class ElementFinder(ContextAware):

    def __init__(self, ctx):
        ContextAware.__init__(self, ctx)
        strategies = {
            'identifier': self._find_by_identifier,
            'id': self._find_by_id,
            'name': self._find_by_name,
            'xpath': self._find_by_xpath,
            'dom': self._find_by_dom,
            'link': self._find_by_link_text,
            'partial link': self._find_by_partial_link_text,
            'css': self._find_by_css_selector,
            'class': self._find_by_class_name,
            'jquery': self._find_by_sizzle_selector,
            'sizzle': self._find_by_sizzle_selector,
            'tag': self._find_by_tag_name,
            'scLocator': self._find_by_sc_locator,
            'default': self._find_by_default
        }
        self._strategies = NormalizedDict(initial=strategies, caseless=True,
                                          spaceless=True)
        self._default_strategies = list(strategies)
        self._key_attrs = {
            None: ['@id', '@name'],
            'a': ['@id', '@name', '@href',
                  'normalize-space(descendant-or-self::text())'],
            'img': ['@id', '@name', '@src', '@alt'],
            'input': ['@id', '@name', '@value', '@src'],
            'button': ['@id', '@name', '@value',
                       'normalize-space(descendant-or-self::text())']
        }

    def find(self, locator, tag=None, first_only=True, required=True,
             parent=None):
        element_type = 'Element' if not tag else tag.capitalize()
        if parent and not self._is_webelement(parent):
            raise ValueError('Parent must be Selenium WebElement but it '
                             'was {}.'.format(type(parent)))
        if self._is_webelement(locator):
            return locator
        prefix, criteria = self._parse_locator(locator)
        strategy = self._strategies[prefix]
        tag, constraints = self._get_tag_and_constraints(tag)
        elements = strategy(criteria, tag, constraints,
                            parent=parent or self.driver)
        if required and not elements:
            raise ElementNotFound("%s with locator '%s' not found."
                                  % (element_type, locator))
        if first_only:
            if not elements:
                return None
            return elements[0]
        return elements

    def register(self, strategy_name, strategy_keyword, persist=False):
        strategy = CustomLocator(self.ctx, strategy_name, strategy_keyword)
        if strategy.name in self._strategies:
            raise RuntimeError("The custom locator '%s' cannot be registered. "
                               "A locator of that name already exists."
                               % strategy.name)
        self._strategies[strategy.name] = strategy.find
        if is_falsy(persist):
            # Unregister after current scope ends
            events.on('scope_end', 'current', self.unregister, strategy.name)

    def unregister(self, strategy_name):
        if strategy_name in self._default_strategies:
            raise RuntimeError("Cannot unregister the default strategy '%s'."
                               % strategy_name)
        if strategy_name not in self._strategies:
            raise RuntimeError("Cannot unregister the non-registered strategy '%s'."
                               % strategy_name)
        del self._strategies[strategy_name]

    def _is_webelement(self, element):
        # Hook for unit tests
        return isinstance(element, WebElement)

    def _disallow_webelement_parent(self, element):
        if self._is_webelement(element):
            raise ValueError('This method does not allow WebElement as parent')

    def _find_by_identifier(self, criteria, tag, constraints, parent):
        elements = self._normalize(parent.find_elements_by_id(criteria)) \
            + self._normalize(parent.find_elements_by_name(criteria))
        return self._filter_elements(elements, tag, constraints)

    def _find_by_id(self, criteria, tag, constraints, parent):
        return self._filter_elements(parent.find_elements_by_id(criteria),
                                     tag, constraints)

    def _find_by_name(self, criteria, tag, constraints, parent):
        return self._filter_elements(parent.find_elements_by_name(criteria),
                                     tag, constraints)

    def _find_by_xpath(self, criteria, tag, constraints, parent):
        return self._filter_elements(parent.find_elements_by_xpath(criteria),
                                     tag, constraints)

    def _find_by_dom(self, criteria, tag, constraints, parent):
        self._disallow_webelement_parent(parent)
        result = self.driver.execute_script("return %s;" % criteria)
        if result is None:
            return []
        if not isinstance(result, list):
            result = [result]
        return self._filter_elements(result, tag, constraints)

    def _find_by_sizzle_selector(self, criteria, tag, constraints, parent):
        self._disallow_webelement_parent(parent)
        js = "return jQuery('%s').get();" % criteria.replace("'", "\\'")
        return self._filter_elements(
            self.driver.execute_script(js),
            tag, constraints)

    def _find_by_link_text(self, criteria, tag, constraints, parent):
        return self._filter_elements(
            parent.find_elements_by_link_text(criteria),
            tag, constraints)

    def _find_by_partial_link_text(self, criteria, tag, constraints, parent):
        return self._filter_elements(
            parent.find_elements_by_partial_link_text(criteria),
            tag, constraints)

    def _find_by_css_selector(self, criteria, tag, constraints, parent):
        return self._filter_elements(
            parent.find_elements_by_css_selector(criteria),
            tag, constraints)

    def _find_by_class_name(self, criteria, tag, constraints, parent):
        return self._filter_elements(
            parent.find_elements_by_class_name(criteria),
            tag, constraints)

    def _find_by_tag_name(self, criteria, tag, constraints, parent):
        return self._filter_elements(
            parent.find_elements_by_tag_name(criteria),
            tag, constraints)

    def _find_by_sc_locator(self, criteria, tag, constraints, parent):
        self._disallow_webelement_parent(parent)
        js = "return isc.AutoTest.getElement('%s')" % criteria.replace("'", "\\'")
        return self._filter_elements([self.driver.execute_script(js)],
                                     tag, constraints)

    def _find_by_default(self, criteria, tag, constraints, parent):
        if tag in self._key_attrs:
            key_attrs = self._key_attrs[tag]
        else:
            key_attrs = self._key_attrs[None]
        xpath_criteria = escape_xpath_value(criteria)
        xpath_tag = tag if tag is not None else '*'
        xpath_constraints = self._get_xpath_constraints(constraints)
        xpath_searchers = ["%s=%s" % (attr, xpath_criteria) for attr in key_attrs]
        xpath_searchers.extend(self._get_attrs_with_url(key_attrs, criteria))
        xpath = "//%s[%s%s(%s)]" % (
            xpath_tag,
            ' and '.join(xpath_constraints),
            ' and ' if xpath_constraints else '',
            ' or '.join(xpath_searchers)
        )
        return self._normalize(parent.find_elements_by_xpath(xpath))

    def _get_xpath_constraints(self, constraints):
        xpath_constraints = [self._get_xpath_constraint(name, value)
                             for name, value in constraints.items()]
        return xpath_constraints

    def _get_xpath_constraint(self, name, value):
        if isinstance(value, list):
            return "@%s[. = '%s']" % (name, "' or . = '".join(value))
        else:
            return "@%s='%s'" % (name, value)

    def _get_tag_and_constraints(self, tag):
        if tag is None:
            return None, {}
        tag = tag.lower()
        constraints = {}
        if tag == 'link':
            tag = 'a'
        if tag == 'partial link':
            tag = 'a'
        elif tag == 'image':
            tag = 'img'
        elif tag == 'list':
            tag = 'select'
        elif tag == 'radio button':
            tag = 'input'
            constraints['type'] = 'radio'
        elif tag == 'checkbox':
            tag = 'input'
            constraints['type'] = 'checkbox'
        elif tag == 'text field':
            tag = 'input'
            constraints['type'] = ['date', 'datetime-local', 'email', 'month',
                                   'number', 'password', 'search', 'tel',
                                   'text', 'time', 'url', 'week', 'file']
        elif tag == 'file upload':
            tag = 'input'
            constraints['type'] = 'file'
        elif tag == 'text area':
            tag = 'textarea'
        return tag, constraints

    def _parse_locator(self, locator):
        if locator.startswith(('//', '(//')):
            return 'xpath', locator
        index = self._get_locator_separator_index(locator)
        if index != -1:
            prefix = locator[:index].strip()
            if prefix in self._strategies:
                return prefix, locator[index+1:].lstrip()
        return 'default', locator

    def _get_locator_separator_index(self, locator):
        if '=' not in locator:
            return locator.find(':')
        if ':' not in locator:
            return locator.find('=')
        return min(locator.find('='), locator.find(':'))

    def _element_matches(self, element, tag, constraints):
        if not element.tag_name.lower() == tag:
            return False
        for name in constraints:
            if isinstance(constraints[name], list):
                if element.get_attribute(name) not in constraints[name]:
                    return False
            elif element.get_attribute(name) != constraints[name]:
                return False
        return True

    def _filter_elements(self, elements, tag, constraints):
        elements = self._normalize(elements)
        if tag is None:
            return elements
        return [element for element in elements
                if self._element_matches(element, tag, constraints)]

    def _get_attrs_with_url(self, key_attrs, criteria):
        attrs = []
        url = None
        xpath_url = None
        for attr in ['@src', '@href']:
            if attr in key_attrs:
                if url is None or xpath_url is None:
                    url = self._get_base_url() + "/" + criteria
                    xpath_url = escape_xpath_value(url)
                attrs.append("%s=%s" % (attr, xpath_url))
        return attrs

    def _get_base_url(self):
        url = self.driver.current_url
        if '/' in url:
            url = '/'.join(url.split('/')[:-1])
        return url

    def _normalize(self, elements):
        # Apparently IEDriver has returned invalid data earlier and recently
        # ChromeDriver has done sometimes returned None:
        # https://github.com/SeleniumHQ/selenium/issues/4555
        if not isinstance(elements, list):
            logger.debug("WebDriver find returned %s" % elements)
            return []
        return elements



# Copyright 2019-     DNB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from .RequestsKeywords import RequestsKeywords
from .version import __version__


class rbfHTTP(RequestsKeywords):
    """ rbfHTTP is a HTTP client library

        Examples:
        
        *GET Requests*
        
        | `Create Session` | google | http://www.google.com |
        | `Create Session` | github | https://api.github.com | verify=${CURDIR}${/}cacert.pem |
        | ${resp}= | `Get Request` | google |  / |
        | `Should Be Equal As Strings` | ${resp.status_code} | 200 |
        | ${resp}= | `Get Request` | github | /users/bulkan |
        | `Should Be Equal As Strings` |  ${resp.status_code} | 200 |
        | `Dictionary Should Contain Value` | ${resp.json()} | Bulkan Evcimen |
        
        *GET Request with URL Parameters*
        
        | `Create Session` | httpbin | http://httpbin.org  |  
        | &{params}= | `Create Dictionary` | key=value | key2=value2 |  
        | ${resp}= | `Get Request` | httpbin | /get | params=${params} |  
        | `Should Be Equal As Strings` | ${resp.status_code} | 200 |  
        | ${jsondata}= | `To Json` | ${resp.content} |  
        | `Should Be Equal` | ${jsondata['args']} | ${params} |  
        
        *GET Request with JSON Data*
        
        | `Create Session` | httpbin | http://httpbin.org |
        | &{data}= | `Create Dictionary` |  latitude=30.496346 | longitude=-87.640356 |  
        | ${resp}= | `Get Request` | httpbin | /get | json=${data} |
        | `Should Be Equal As Strings` | ${resp.status_code} | 200 |
        | ${jsondata}= | `To Json` | ${resp.content} |
        
        *GET Request (HTTPS) and verify cert*
        
        | `Create Session` | httpbin | https://httpbin.org | verify=True |
        | ${resp}= | `Get Request` | httpbin | /get |
        | `Should Be Equal As Strings` | ${resp.status_code} | 200 |
        
        *GET Request (HTTPS) and verify cert with CA bundle*
        
        | `Create Session` | httpbin | https://httpbin.org |  verify=${CURDIR}${/}cacert.pem |
        | ${resp}= | `Get Request` | httpbin | /get |
        | `Should Be Equal As Strings` | ${resp.status_code} | 200 |
        
        *GET Request (HTTPS) with client side certificates*
        
        | @{client_certs}= | `Create List` | ${CURDIR}${/}clientcert.pem | ${CURDIR}${/}clientkey.pem |
        | `Create Client Cert Session` | crtsession | https://server.cryptomix.com/secure | client_certs=@{client_certs} |
        | ${resp}= | `Get Request` | crtsession | / |
        | `Should Be Equal As Strings` | ${resp.status_code} | 200 |
        
        *GET Request with Auth*
        
        | ${auth}= | `Create List` | user |  passwd |
        | `Create Session` | httpbin | https://httpbin.org | auth=${auth} | verify=${CURDIR}${/}cacert.pem |
        | ${resp}= | `Get Request` | httpbin | /basic-auth/user/passwd |
        | `Should Be Equal As Strings` | ${resp.status_code} | 200 |
        | `Should Be Equal As Strings` | ${resp.json()['authenticated']} | True |
        
        *GET Request with Custom Auth*
        
        | ${auth}= | `Get Custom Auth` | user |  passwd | 
        | `Create Custom Session` | httpbin | https://httpbin.org | auth=${auth} | verify=${CURDIR}${/}cacert.pem | 
        | ${resp}= | `Get Request` | httpbin | /basic-auth/user/passwd | 
        | `Should Be Equal As Strings` | ${resp.status_code} | 200 | 
        | `Should Be Equal As Strings` | ${resp.json()['authenticated']} | True | 
        
        *GET Request with Digest Auth*
        
        | ${auth}= | `Create List` | user | pass | 
        | `Create Digest Session` | httpbin | https://httpbin.org | auth=${auth} | debug=3 | verify=${CURDIR}${/}cacert.pem | 
        | ${resp}= | `Get Request` | httpbin | /digest-auth/auth/user/pass | 
        | `Should Be Equal As Strings` | ${resp.status_code} | 200 | 
        | `Should Be Equal As Strings` | ${resp.json()['authenticated']} | True | 
        
        *GET Request with Redirection*
        
        | `Create Session` | httpbin | http://httpbin.org | debug=3 |
        | ${resp}= | `Get Request` | httpbin | /redirect/1 |
        | `Should Be Equal As Strings` | ${resp.status_code} | 200 |
        | ${resp}= | `Get Request` | httpbin | /redirect/1 | allow_redirects=${true} |
        | `Should Be Equal As Strings` | ${resp.status_code} | 200 |
        
        *GET Request without Redirection*
        
        | `Create Session` | httpbin | http://httpbin.org |
        | ${resp}= | `Get Request` | httpbin | /redirect/1 | allow_redirects=${false} |
        | ${status}= | `Convert To String` | ${resp.status_code} |
        | `Should Start With` | ${status} | 30 |        
        
        *POST Requests*
        
        | `Create Session` | httpbin | http://httpbin.org |
        | &{data}= | `Create Dictionary` | name=bulkan | surname=evcimen |
        | &{headers}= | `Create Dictionary` | Content-Type=application/x-www-form-urlencoded |
        | ${resp}= | `Post Request` | httpbin | /post | data=${data} | headers=${headers} |
        | `Dictionary Should Contain Value` | ${resp.json()['form']} | bulkan |
        | `Dictionary Should Contain Value` | ${resp.json()['form']} | evcimen |
                
        *POST Request with File*
        
        | `Create Session` | httpbin | http://httpbin.org |
        | ${file_data}= | `Get Binary File` | ${CURDIR}${/}data.json |
        | &{files}= | `Create Dictionary` | file=${file_data} |
        | ${resp}= | `Post Request` | httpbin | /post | files=${files} |
        | ${file}= | `To Json` | ${resp.json()['files']['file']} |
        | `Dictionary Should Contain Key` | ${file} | one |
        | `Dictionary Should Contain Key` | ${file} | two | 
        
        *POST Request with Data and File*
        
        | `Create Session` | httpbin | http://httpbin.org |
        | &{data}= | `Create Dictionary` | name=mallikarjunarao | surname=kosuri |
        | `Create File` | foobar.txt | content=foobar |
        | ${file_data}= | `Get File` | foobar.txt |
        | &{files}= | `Create Dictionary` | file=${file_data} |
        | ${resp}= | `Post Request` | httpbin | /post | files=${files} | data=${data} |
        | `Should Be Equal As Strings` |  ${resp.status_code} | 200 |
        
        *POST Request with URL Params*
        
        | `Create Session` | httpbin | http://httpbin.org |
        | &{params}= |  `Create Dictionary` |  key=value | key2=value2 |
        | ${resp}= | `Post Request` | httpbin | /post    | params=${params} |
        | `Should Be Equal As Strings` | ${resp.status_code} | 200 |

        *POST Request with JSON Data*
        
        | `Create Session` | httpbin | http://httpbin.org | 
        | &{data}= | `Create Dictionary` |  latitude=30.496346 | longitude=-87.640356 |  
        | ${resp}= | `Post Request` | httpbin | /post | json=${data} | 
        | `Should Be Equal As Strings` | ${resp.status_code} | 200 | 
        | ${jsondata}= | `To Json` | ${resp.content} | 
        | `Should Be Equal` | ${jsondata['json']} | ${data} | 
        
        *POST Request with NO Data*
        
        | `Create Session` | httpbin | http://httpbin.org |
        | ${resp}= | `Post Request` | httpbin | /post | 
        | `Should Be Equal As Strings` | ${resp.status_code} | 200 |
               
        *POST Request With No Dictionary*
        
        | `Set Test Variable` | ${data} | some content |
        | ${resp}= | `Post Request` | httpbin | /post | data=${data} |
        | `Should Be Equal As Strings` | ${resp.status_code} | 200 |
        | `Should Contain` | ${resp.text} | ${data} |
        
        *POST Request with Unicode Data*
        
        | `Create Session` | httpbin | http://httpbin.org | debug=3 |
        | &{data}= | `Create Dictionary` | name=åº¦å�‡æ�‘ |
        | &{headers}= | `Create Dictionary` | Content-Type=application/x-www-form-urlencoded |
        | ${resp}= | `Post Request` | httpbin | /post | data=${data} | headers=${headers} |
        | `Dictionary Should Contain Value` | ${resp.json()['form']} | åº¦å�‡æ�‘ |
        
        *POST Request with Binary Data in Dictionary*
        
        | `Create Session` | httpbin | http://httpbin.org | debug=3 |
        | ${file_data}= | `Get Binary File` | ${CURDIR}${/}data.json |
        | &{data}= | `Create Dictionary` | name=${file_data.strip()} |
        | &{headers}= | `Create Dictionary` | Content-Type=application/x-www-form-urlencoded |
        | ${resp}= | `Post Request` | httpbin | /post | data=${data} | headers=${headers} |
        | `Log` | ${resp.json()['form']} |
        | `Should Contain` | ${resp.json()['form']['name']} | \u5ea6\u5047\u6751 |
        
        *POST Request with Binary Data*
        
        | `Create Session` | httpbin | http://httpbin.org | debug=3 |
        | ${data}= | `Get Binary File` | ${CURDIR}${/}data.json |
        | &{headers}= | `Create Dictionary` | Content-Type=application/x-www-form-urlencoded |
        | ${resp}= | `Post Request` | httpbin | /post | data=${data} | headers=${headers} |
        | `Log` | ${resp.json()['form']} |
        | ${value}= | `evaluate` | list(${resp.json()}['form'].keys())[0] |
        | `Should Contain` | ${value} | åº¦å�‡æ�‘ |
        
        *POST Request with Arbitrary Binary Data*
        
        | `Create Session` | httpbin | http://httpbin.org | debug=3 | 
        | ${data}= | `Get Binary File` | ${CURDIR}${/}randombytes.bin | 
        | &{headers}= | `Create Dictionary` | Content-Type=application/octet-stream |  Accept=application/octet-stream | 
        | ${resp}= | `Post Request` | httpbin | /post | data=${data} | headers=&{headers} | 
        | # TODO Compare binaries. Content is json with base64 encoded data | 
        | `Log` | "Success" |
        
        *POST request with Redirection*
        
        | `Create Session` | jigsaw | http://jigsaw.w3.org |
        | ${resp}= | `Post Request` | jigsaw | /HTTP/300/302.html |
        | `Should Be Equal As Strings` | ${resp.status_code} | 200 |
        | ${resp}= | `Post Request` | jigsaw | /HTTP/300/302.html | allow_redirects=${true} |
        | `Should Be Equal As Strings` | ${resp.status_code} | 200 |
        
        *POST request without Redirection*
        
        | `Create Session` | jigsaw | http://jigsaw.w3.org | debug=3 |
        | ${resp}= | `Post Request` | jigsaw | /HTTP/300/302.html | allow_redirects=${false} |
        | ${status}= | `Convert To String` | ${resp.status_code} |
        | `Should Start With` | ${status} | 30 |
        
        *PUT Request*
        
        | `Create Session` | httpbin | http://httpbin.org |
        | &{data}= | `Create Dictionary` | name=bulkan | surname=evcimen |
        | &{headers}= | `Create Dictionary` | Content-Type=application/x-www-form-urlencoded | 
        | ${resp}= | `Put Request` | httpbin | /put | data=${data} | headers=${headers} |
        | `Dictionary Should Contain Value` | ${resp.json()['form']} | bulkan |
        | `Dictionary Should Contain Value` | ${resp.json()['form']} | evcimen |
        
        *PUT Request with URL Params*
        
        | `Create Session` | httpbin | http://httpbin.org |
        | &{params}= |  `Create Dictionary` |  key=value | key2=value2 |
        | ${resp}= | `Put Request` | httpbin | /put | params=${params} |
        | `Should Be Equal As Strings` | ${resp.status_code} | 200 |
                
        *PUT Request with JSON Data *
        
        | `Create Session` | httpbin | http://httpbin.org |
        | &{data}= | `Create Dictionary` | latitude=30.496346 | longitude=-87.640356 |
        | ${resp}= | `Put Request` | httpbin | /put | json=${data} |
        | `Should Be Equal As Strings` | ${resp.status_code} | 200 |
        | ${jsondata}= | `To Json` | ${resp.content} |
        | `Should Be Equal` |  ${jsondata['json']} | ${data} |
        
        *PUT Request with No Dictionary*
        
        | `Create Session` | httpbin | http://httpbin.org |
        | `Set Test Variable` | ${data} | some content |
        | ${resp}= | `Put Request` | httpbin | /put | data=${data} |
        | `Should Be Equal As Strings` | ${resp.status_code} | 200 |
        | `Should Contain` | ${resp.text} | ${data} |
        
        *PUT Request with NO Data*
        
        | `Create Session` | httpbin | http://httpbin.org |
        | ${resp}= | `Put Request` | httpbin | /put |
        | `Should Be Equal As Strings` | ${resp.status_code} | 200 |
        
        *PUT request with Redirection*
        
        | `Create Session` | jigsaw | http://jigsaw.w3.org | debug=3 |
        | ${resp}= | `Put Request` | jigsaw | /HTTP/300/302.html |
        | `Should Be Equal As Strings` | ${resp.status_code} | 200 |
        | ${resp}= | `Put Request` | jigsaw | /HTTP/300/302.html | allow_redirects=${true} | 
        | `Should Be Equal As Strings` | ${resp.status_code} | 200 |
        
        *PUT request without Redirection*
        
        | `Create Session` | jigsaw | http://jigsaw.w3.org |
        | ${resp}= | `Put Request` | jigsaw | /HTTP/300/302.html | allow_redirects=${false} |
        | ${status}= | `Convert To String` | ${resp.status_code} |
        | `Should Start With` | ${status} | 30 |
        
        *HEAD Request*
        
        | `Create Session` | httpbin | http://httpbin.org |
        | ${resp}= | `Head Request` | httpbin | /headers |
        | `Should Be Equal As Strings` | ${resp.status_code} | 200 |
        
        *HEAD request with Redirection*
        
        | `Create Session` | httpbin | http://httpbin.org |
        | ${resp}= | `Head Request` | httpbin | /redirect/1 | allow_redirects=${true} |
        | `Should Be Equal As Strings` | ${resp.status_code} | 200 |
        
        *HEAD request without Redirection*
        
        | `Create Session` | httpbin | http://httpbin.org |
        | ${resp}= | `Head Request` | httpbin | /redirect/1 |
        | ${status}= | `Convert To String` | ${resp.status_code} |
        | `Should Start With` | ${status} | 30 |
        | ${resp}= | `Head Request` | httpbin | /redirect/1 | allow_redirects=${false} |
        | ${status}= | `Convert To String` | ${resp.status_code} |
        | `Should Start With` | ${status} | 30 |
        
        *OPTIONS Request*
        
        | `Create Session` | httpbin | http://httpbin.org |
        | ${resp}= | `Options Request` | httpbin | /headers |
        | `Should Be Equal As Strings` | ${resp.status_code} | 200 |
        | `Dictionary Should Contain Key` | ${resp.headers} | allow |
        
        *OPTIONS request with Redirection*
        
        | `Create Session` | httpbin | http://httpbin.org |
        | ${resp}= | `Options Request` | httpbin | /redirect/1 |
        | `Should Be Equal As Strings` | ${resp.status_code} | 200 |
        | ${resp}= | `Options Request` | httpbin | /redirect/1 | allow_redirects=${true} |
        | `Should Be Equal As Strings` | ${resp.status_code} | 200 |  
        
        *DELETE Request with URL Params*
        
        | `Create Session` | httpbin | http://httpbin.org |
        | &{params}= |  `Create Dictionary` |  key=value | key2=value2 |
        | ${resp}= | `Delete Request` | httpbin | /delete    | ${params} |
        | `Should Be Equal As Strings` | ${resp.status_code} | 200 |
        
        *DELETE Request with NO data*
        
        | `Create Session` | httpbin | http://httpbin.org |
        | ${resp}= | `Delete Request` | httpbin | /delete |
        | `Should Be Equal As Strings` | ${resp.status_code} | 200 |
        
        *DELETE Request with Data*
        
        | `Create Session` | httpbin | http://httpbin.org | debug=3 |
        | &{data}= | `Create Dictionary` | name=bulkan | surname=evcimen |
        | ${resp}= | `Delete Request` | httpbin | /delete | data=${data} |
        | `Should Be Equal As Strings` | ${resp.status_code} | 200 |
        | `Log` | ${resp.content} |
        | Comment | `Dictionary Should Contain Value` | ${resp.json()['data']} | bulkan |
        | Comment | `Dictionary Should Contain Value` | ${resp.json()['data']} | evcimen |
        
        *DELETE Request with JSON Data*
        
        | `Create Session` | httpbin | http://httpbin.org |
        | &{data}= | `Create Dictionary` |  latitude=30.496346 | longitude=-87.640356 |
        | ${resp}= | `Delete Request` | httpbin | /delete | json=${data} |
        | `Should Be Equal As Strings` | ${resp.status_code} | 200 |
        | ${jsondata}= | `To Json` | ${resp.content} |
        | `Should Be Equal` | ${jsondata['json']} | ${data} |
        
        *PATCH Requests*
        
        | `Create Session`  |  httpbin  |  http://httpbin.org |
        | &{data}=  |  `Create Dictionary`  |  name=bulkan  |  surname=evcimen |
        | &{headers}=  |  `Create Dictionary`  |  Content-Type=application/x-www-form-urlencoded |
        | ${resp}=  |  `Patch Request`  |  httpbin  |  /patch  |  data=${data}  |  headers=${headers} |
        | `Dictionary Should Contain Value`  |  ${resp.json()['form']}  |  bulkan |
        | `Dictionary Should Contain Value`  |  ${resp.json()['form']}  |  evcimen |
        
        *PATCH Requests with JSON Data*
        
        | `Create Session` | httpbin  |   http://httpbin.org |
        | &{data}=  |  `Create Dictionary` |  latitude=30.496346 | longitude=-87.640356 |
        | ${resp}=  |   `Patch Request` | httpbin | /patch  |  json=${data} |
        | `Should Be Equal As Strings` | ${resp.status_code} | 200 |
        | ${jsondata}= | `To Json` | ${resp.content} |
        | `Should Be Equal`  |   ${jsondata['json']}  |   ${data} | 
 
        *Pretty Print JSON Object*
        
        | `Log`  |  ${resp} |
        | ${output}=  |  `To Json`  |  ${resp.content}  |  pretty_print=True |
        | `Log`  |  ${output} |
        | `Should Contain`  |  ${output}  |  "key_one": "true" |
        | `Should Contain`  |  ${output}  |  "key_two": "this is a test string" |
        | `Should Not Contain`  |  ${output}  |  {u'key_two': u'this is a test string', u'key_one': u'true'} |

    """
    ROBOT_LIBRARY_SCOPE = 'GLOBAL'
    ROBOT_LIBRARY_VERSION = __version__






# Copyright 2019-     DNB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings

from robot.api import logger
from robot.libraries.BuiltIn import BuiltIn

from .rbfFTP import rbfFTP
from .version import __version__



class rbfFTP(rbfFTP):
    """rbfFTP is a ftp client library for Robot Framework.

    This library provides functionality of FTP client.


The simplest example (connect, change working dir, print working dir, close):
 | ftp connect | 192.168.1.10 | mylogin | mypassword |
 | cwd | /home/myname/tmp/testdir |
 | pwd |
 | ftp close |

It is possible to use multiple ftp connections in parallel. Connections are
identified by string identifiers:
 | ftp connect | 192.168.1.10 | mylogin | mypassword | connId=ftp1 |
 | ftp connect | 192.168.1.20 | mylogin2 | mypassword2 | connId=ftp2 |
 | cwd | /home/myname/tmp/testdir | ftp1 |
 | cwd | /home/myname/tmp/testdir | ftp2 |
 | pwd | ftp2 |
 | pwd | ftp1 |
 | ftp close | ftp2 |
 | ftp close | ftp1 |

To run library remotely execute: python rbfFTP.py <ipaddress> <portnumber>
(for example: python rbfFTP.py 192.168.0.101 8222)
    
    """
    ROBOT_LIBRARY_SCOPE = 'TEST SUITE'
    ROBOT_LIBRARY_VERSION = __version__
